
from ... classes.flux_row_cls import flux_row_cls


# noinspection PyMissingConstructor
class lev_row_cls(flux_row_cls):
    @classmethod
    def reserved_names(cls):
        return ['address'] + super().reserved_names()

    def __init__(self, headers, values, address=''):
        self.__dict__['_headers'] = headers
        self.__dict__['values']   = values
        self.__dict__['address']  = address






import gc

from array import array
from collections import Counter
from collections import namedtuple
from collections import ItemsView
from copy import deepcopy

from typing import Generator
from typing import List
from typing import Tuple
from typing import Dict
from typing import Union
from typing import Any

from .flux_row_cls import flux_row_cls

from ..util.filesystem import parse_file_extension
from ..util.filesystem import read_file
from ..util.filesystem import write_file
from ..util.filesystem import pickle_extensions
from ..util.filesystem import json_dumps_extended

from ..util.iter import IterationDepthError
from ..util.iter import ColumnNameError
from ..util.iter import base_class_names
from ..util.iter import is_exhaustable
from ..util.iter import is_collection
from ..util.iter import is_subscriptable
from ..util.iter import standardize_variable_arity_values
from ..util.iter import iteration_depth
from ..util.iter import iterator_to_collection
from ..util.iter import map_values_to_enum
from ..util.iter import are_indices_contiguous
from ..util.iter import modify_iteration_depth
from ..util.iter import transpose
from ..util.iter import to_grouped_dict

from ..util.text import print_runtime
from ..util.text import deprecated
from ..util.text import object_name
from ..util.text import format_header
from ..util.text import format_header_lite
from ..util.text import format_integer
from ..util.text import function_parameters
from ..util.text import function_name
from ..util.text import vengeance_message

from ..util.classes.namespace_cls import namespace_cls

from ..conditional import ordereddict
from ..conditional import line_profiler_installed
from ..conditional import numpy_installed

if numpy_installed:
    import numpy


class flux_cls:
    """ primary data management class

    * converts a list of lists (a matrix) to a list of flux_row_cls objects
    * similar to a pandas DataFrame, but optimized for row-major ordered operations (df.iterrows())

    eg:
        # first row used as header names (like a csv)
        matrix = [['col_a', 'col_b', 'col_c'],
                  ['a', 'b', 'c']]
        flux = flux_cls(matrix)
        flux = flux_cls.from_csv('file.csv')

        for row in flux:
            a = row.col_a
            row.col_a = None
            
        flux['col_d'] = [o.upper() for c in flux['col_a']]
        flux.sort('col_a', 'col_b')
        flux.filter(lambda: row: row.col_a == 'value')
        
        flux.to_csv('file.csv')
        flux = flux_cls.from_csv('file.csv')
    """

    # as_array_preview_indices
    aap_indices = [1, 5 + 1]

    def __init__(self, matrix=None):
        self.headers: Dict[Union[str, bytes], int]
        self.matrix:  List[flux_row_cls]

        gc_enabled = gc.isenabled()
        gc.disable()

        matrix  = self.__validate_matrix(matrix)
        headers = self.__validate_names_as_headers(matrix[0])
        matrix  = [flux_row_cls(headers, row) for row in matrix]

        if gc_enabled: gc.enable()

        self.headers = headers
        self.matrix  = matrix

    @property
    def num_cols(self):
        return len(self.matrix[0].values)

    @property
    def num_rows(self):
        """ header row is not included, compare to self.__len__ """
        return len(self.matrix) - 1

    @property
    def as_array_preview(self):
        """
        to help with debugging: meant to trigger a debugging feature in PyCharm
        PyCharm will recognize the numpy array and enable the "...view as array"
        option in the debugger which displays values in a special window as a table

        self.aap_indices are class attributes
        """
        return self.as_array(self.aap_indices[0], self.aap_indices[1])

    def as_array(self, r_1=1, r_2=None):
        """
        to help with debugging: meant to trigger a debugging feature in PyCharm
        PyCharm will recognize the numpy array and enable the "...view as array"
        option in the debugger which displays values in a special window as a table
        """
        # region {closure functions}
        def start_row_index(r):
            if r is None:
                return 1
            elif r < 0:
                return len(self.matrix) + r
            else:
                return r

        def stop_row_index(r):
            if r is None:
                return len(self.matrix)
            elif r < 0:
                return len(self.matrix) + r
            else:
                return r

        def format_row_index(i):
            i = format_integer(i)
            i = format_header_lite(i)

            return i
        # endregion

        r_1 = max(start_row_index(r_1), 1)
        r_2 = min(stop_row_index(r_2), len(self.matrix))
        if r_1 > r_2:
            r_1, r_2 = r_2, r_1

        n_c  = self.num_cols
        rows = self.matrix[r_1:r_2]
        if rows:
            c_m = max(n_c, *[len(row.values) for row in rows])
            has_label = ('r_i' in rows[0].__dict__)
        else:
            c_m = n_c
            has_label = False

        is_jagged = (c_m != n_c)

        h_i = format_header_lite('address')
        h_v = [format_header(n) for n in self.header_names()]
        h_j = ['🗲'] * (c_m - len(h_v))

        if has_label:
            h_v.insert(0, format_header_lite('r_i'))
            c_m += 1

        m = [[h_i, *h_v, *h_j]]
        for r_i, row in enumerate(rows, r_1):
            r_i = format_row_index(r_i)

            if has_label:
                r_v = [format_row_index(row.__dict__.get('r_i', -1)),
                       *row.values]
            else:
                r_v = row.values

            if is_jagged:
                r_j = ['🗲jagged🗲'] * (c_m - len(r_v))
                m.append([r_i, *r_v, *r_j])
            else:
                m.append([r_i, *r_v])

        if not numpy_installed:
            m = '\n'.join([str(row) for row in m])
            return m

        return numpy.array(m, dtype=object)

    def header_names(self):
        """
        self.matrix[0].values and self.headers.keys may not always be identical
            map_values_to_enum() makes certain modifications to self.headers.keys,
            such as coercing values to strings, modifying duplicate values, etc
        """
        return list(self.headers.keys())

    def is_empty(self) -> bool:
        for row in self.matrix:
            if row.values:
                return False

        return True

    def has_duplicate_row_pointers(self) -> bool:
        """
        if row.values share pointers to the same underlying list,
        this will usually cause unwanted behaviors to any modifications of row.values
        """
        row_ids = set()

        for row in self.matrix:
            row_id = id(row.values)
            if row_id in row_ids:
                return True

            row_ids.add(row_id)

        return False

    def duplicate_row_pointers(self) -> Dict:
        enumrow_nt = namedtuple('EnumRow', ('i', 'row'))

        d = ordereddict()
        for i, row in enumerate(self.matrix):
            row_id = id(row.values)
            row    = enumrow_nt(i, row)

            if row_id in d:
                d[row_id].append(row)
            else:
                d[row_id] = [row]

        d = ordereddict([('\\x{:x}'.format(row_id), rows) for row_id, rows in d.items()
                                                          if len(rows) > 1])
        return d

    def is_jagged(self) -> bool:
        num_cols = len(self.headers)

        for row in self.matrix:
            if len(row.values) != num_cols:
                return True

        return False

    def jagged_rows(self) -> Generator[namedtuple, None, None]:
        enumrow_nt = namedtuple('EnumRow', ('i', 'row'))

        num_cols = self.num_cols
        for i, row in enumerate(self.matrix):
            if len(row.values) != num_cols:
                yield enumrow_nt(i, row)

    # region {filesystem methods}
    def to_file(self, path,
                      encoding=None,
                      filetype=None,
                      **kwargs):

        filetype = parse_file_extension((filetype or path), include_dot=True).lower()

        if filetype == '.csv':
            return self.to_csv(path, encoding, **kwargs)
        elif filetype == '.json':
            return self.to_json(path, encoding, **kwargs)
        elif filetype in pickle_extensions:
            return self.serialize(path, **kwargs)

        raise ValueError("invalid filetype: '{}' \nfiletype must be in {}"
                         .format(filetype, ['.csv', '.json'] + list(pickle_extensions)))

    @classmethod
    def from_file(cls, path,
                       encoding=None,
                       filetype=None,
                       **kwargs):

        filetype = parse_file_extension((filetype or path), include_dot=True).lower()

        if filetype == '.csv':
            return cls.from_csv(path, encoding, **kwargs)
        if filetype == '.json':
            return cls.from_json(path, encoding, **kwargs)
        if filetype in pickle_extensions:
            return cls.deserialize(path, **kwargs)

        raise ValueError("invalid filetype: '{}' \nfiletype must be in {}"
                         .format(filetype, ['.csv', '.json'] + list(pickle_extensions)))

    def to_csv(self, path,
                     encoding=None,
                     **kwargs):

        write_file(path, self.rows(), encoding, filetype='.csv', **kwargs)
        return self

    @classmethod
    def from_csv(cls, path,
                      encoding=None,
                      **kwargs):

        m = read_file(path, encoding, filetype='.csv', **kwargs)
        return cls(m)

    def to_json(self, path=None,
                      encoding=None,
                      **kwargs):

        h = tuple(self.header_names())
        o = [ordereddict(zip(h, row.values)) for row in self.matrix[1:]]

        if path is None:
            return json_dumps_extended(o, **kwargs)

        write_file(path, o, encoding, filetype='.json', **kwargs)
        return self

    @classmethod
    def from_json(cls, path,
                       encoding=None,
                       **kwargs):

        o = read_file(path, encoding, filetype='.json', **kwargs)
        # if isinstance(o, dict):
        #     o = [o]

        return cls(o)

    def serialize(self, path, **kwargs):
        """
        although convenient, pickle introduces significant security flaws
        you should be sure no malicious actors have access to the location of these files
        """
        write_file(path, self, filetype='.flux', **kwargs)
        return self

    @classmethod
    def deserialize(cls, path, **kwargs):
        """
        although convenient, pickle introduces significant security flaws
        you should be sure no malicious actors have access to the location of these files
        """
        return read_file(path, filetype='.flux', **kwargs)
    # endregion

    def columns(self, *names) -> Generator[Any, None, None]:
        """
        return a one-dimensional list if single column name:
            ['a', 'a', 'a' ...] = list(flux.columns('col_a'))

            flux_b['col_a'] = flux_a['col_a']

        return a two-dimensional list if multiple column names:
            [('a', 'a', 'a', ...),
             ('b', 'b', 'b', ...)] = list(flux.columns('col_a', 'col_b'))

        a, b, c = flux.columns('col_a', 'col_b', 'col_c')
        a, b, c = flux['col_a', 'col_b', 'col_c']
        for a, b, c in zip(*flux[:3]):
            pass
        """
        if names == ():
            names = self.header_names()
        else:
            names = self.__validate_names_not_empty(names, depth_offset=-1)

        has_multiple_columns = isinstance(names, slice) or \
                               (is_collection(names) and len(names) > 1)
        rva = self.row_values_accessor(names)
        col = (rva(row) for row in self.matrix[1:])

        if has_multiple_columns:
            col = transpose(col, astype=tuple)

        return col

    def values(self, *names) -> Generator[Any, None, None]:
        """
        return a one-dimensional list if single column name:
            ['a', 'a', 'a', ...] = list(flux.values('col_a'))

        return a two-dimensional list if multiple column names:
            [('a', 'b')
             ('a', 'b'),
             ('a', 'b'), ...] = list(flux.values('col_a', 'col_b'))

        a, b, c = zip(*flux.values('col_a', 'col_b', 'col_c'))
        for a, b, c in flux.values('col_a', 'col_b', 'col_c'):
            pass
        """
        if names == ():
            names = self.header_names()

        rva = self.row_values_accessor(names)
        return (rva(row) for row in self.matrix[1:])

    def rows(self, r_1=0, r_2=None) -> Generator[Union[List, Tuple], None, None]:
        return (row.values for row in self.matrix[r_1:r_2])

    def dictrows(self, r_1=1, r_2=None) -> Generator[Dict, None, None]:
        names = tuple(self.header_names())
        for row in self.matrix[r_1:r_2]:
            yield ordereddict(zip(names, row.values))

    def namedrows(self, r_1=1, r_2=None) -> Generator[namespace_cls, None, None]:
        names = tuple(self.header_names())
        for row in self.matrix[r_1:r_2]:
            yield namespace_cls(**ordereddict(zip(names, row.values)))

    # noinspection PyArgumentList
    def namedtuples(self, r_1=1, r_2=None) -> Generator[namedtuple, None, None]:
        row_nt = namedtuple('Row', self.header_names())
        for row in self.matrix[r_1:r_2]:
            yield row_nt(*row.values)

    def reset_headers(self, names=None):
        """
        simply re-assigning self.headers to a new variable (eg, self.headers = dict())
        will de-reference all flux_row_cls._headers in matrix,  so self.headers
        must be cleared, then updated
        """
        use_existing_headers = (names is None)

        if use_existing_headers:
            names = self.matrix[0].values

        headers = map_values_to_enum(names)
        self.headers.clear()
        self.headers.update(headers)

        if not use_existing_headers:
            self.matrix[0].values = list(headers.keys())

        return self

    def reset_matrix(self, m):
        (self.headers,
         self.matrix) = self.__validate_headers_and_matrix(m, headers=None)

        return self

    def execute_commands(self, commands, profiler=False, print_commands=False):
        command_nt = namedtuple('Command', ('name', 'method', 'args', 'kwargs'))

        profiler = self.__validate_profiler_function(profiler)
        commands = self.__validate_command_methods(commands, command_nt)

        if print_commands:
            s = '{}.execute_commands:'.format(self.__class__.__name__)
            s = vengeance_message(s)
            print(s)

        completed_commands = []
        for i, command in enumerate(commands, 1):
            name   = command.name
            method = command.method
            args   = command.args
            kwargs = command.kwargs

            if print_commands:
                s = '        {}: @{}()'.format(i, function_name(method))
                print(s)

            if profiler:
                method = profiler(method)

            method(*command.args, **command.kwargs)
            completed_commands.append([name,
                                       args,
                                       kwargs])

        if hasattr(profiler, 'print_stats'):
            profiler.print_stats()

        return completed_commands

    def matrix_by_headers(self, *names):
        if self.is_empty():
            raise ValueError('matrix is empty')

        names = standardize_variable_arity_values(names, depth=1)
        if not names:
            return self
        if isinstance(names, dict):
            names = [names]

        headers = self.headers.copy()
        names   = [self.__validate_renamed_or_inserted_column(name, headers) for name in names]

        all_columns  = [row.values for row in self.matrix[1:]]
        all_columns  = list(transpose(all_columns, astype=list))
        empty_column = [None] * self.num_rows

        m = []
        for n in names:
            if n in headers: column = all_columns[headers[n]]
            else:            column = empty_column

            m.append([n] + column)

        m = transpose(m)
        self.reset_matrix(m)

        return self

    def rename_columns(self, old_to_new_mapping):
        if not isinstance(old_to_new_mapping, dict):
            raise TypeError('old_to_new_mapping must be a dictionary')

        names = self.header_names()
        for h_old, h_new in old_to_new_mapping.items():
            i = self.headers[h_old]
            names[i] = h_new

        self.reset_headers(names)

        return self

    def insert_columns(self, *names, after=False):
        """ eg:
            flux.insert_columns((0,  'inserted'))        insert as first column
            flux.insert_columns((3,  'inserted'))        insert column before 4th column
            flux.insert_columns((-1, 'inserted'))        insert column at end

            flux.insert_columns((1, 'inserted_a'),       insert multiple columns before 1st column
                                (1, 'inserted_b'),
                                (1, 'inserted_c'))

            flux.insert_columns(('col_c', 'inserted'))               insert column before column 'col_c'
            flux.insert_columns(('col_c', 'inserted', after=True))   insert column after column 'col_c'

        note on inner / outer loop performance:
            because python struggles on very tight loops,
            it's faster to use the matrix loop as the inner loop

            * faster:
            for i in indices:
                for row in self.matrix[1:]:
                    ...

            * slower:
            for row in self.matrix[1:]:
                for i in indices:
                    ...
        """
        names = standardize_variable_arity_values(names, depth=1)
        if not names:
            return self

        names = self.__validate_inserted_items(names, self.headers)

        after = [bool(_) for _ in standardize_variable_arity_values(after, depth=1)]
        after.extend([False] * (len(names) - len(after)))

        names = list(reversed(names))
        after = list(reversed(after))

        header_names = self.header_names()

        for item, _after_ in zip(names, after):
            before, header = item

            if isinstance(before, int): i = before
            else:                       i = header_names.index(before)
            if _after_:
                i += 1

            header_names.insert(i, header)

        indices = sorted([header_names.index(h) for _, h in names])

        for i in indices:
            for row in self.matrix[1:]:
                row.values.insert(i, None)

        self.reset_headers(header_names)

        return self

    def append_columns(self, *names, values=None):
        """
        # if max([iteration_depth(n) for n in names]) > 2:
        #     raise IterationDepthError('values not submitted as keyword?')

        # if 'values' in values:
        #     values = values.pop('values')

        # nd = iteration_depth(values)

        # is_single_row    = (v_num_rows == 1)        and (v_num_cols == num_rows)

        # if is_single_value:
            # _row_    = [_values_[0][0] for _ in range(num_cols)]
            # _values_ = [_row_          for _ in range(num_rows)]
            # v = _values_[0][0]
            # _values_ = [[v for _ in range(num_cols)]
            #                for _ in range(num_rows)]
        """
        names = standardize_variable_arity_values(names, depth=1)
        if not names:
            return self

        self.__validate_no_duplicate_names(names)
        self.__validate_no_names_intersect_with_headers(names, self.headers)

        _values_ = standardize_variable_arity_values(values, depth=2)

        num_rows   = len(self.matrix) - 1
        num_cols   = len(names)
        v_num_rows = len(_values_)
        v_num_cols = len(_values_[0])

        mismatched_dimensions = (v_num_rows != num_rows)  or (v_num_cols != num_cols)
        is_single_value       = (v_num_rows == 1)        and (v_num_cols == 1)
        is_single_column      = (v_num_rows == num_rows) and (num_cols == 1)
        is_transposed         = (v_num_rows == num_cols) and (v_num_cols == num_rows)

        if mismatched_dimensions:
            if is_single_value:
                _row_    = [_values_[0][0] for _ in range(num_cols)]
                _values_ = [_row_          for _ in range(num_rows)]
            elif is_single_column:
                _values_ = transpose([_values_], list)
            elif is_transposed:
                _values_ = transpose(_values_, list)
            else:
                raise IndexError('invalid dimensions for column values\n'
                                 'expected: {:,} cols x {:,} rows\n'
                                 'values:   {:,} cols x {:,} rows'.format(num_cols, num_rows, v_num_cols, v_num_rows))

        header_names = self.header_names() + list(names)
        for row, v in zip(self.matrix[1:], _values_):
            row.values.extend(v)

        self.reset_headers(header_names)

        return self

    def delete_columns(self, *names):
        """
        method will fail if columns are jagged, do a try / except on del row.values[i]
        then length check on row values if failure?

        note on inner / outer loop performance:
            because python struggles on very tight loops,
            it's faster to use the matrix loop as the inner loop

            * faster:
            for i in indices:
                for row in self.matrix[1:]:
                    ...

            * slower:
            for row in self.matrix[1:]:
                for i in indices:
                    ...
        """
        names = standardize_variable_arity_values(names, depth=1)
        if not names:
            return self

        indices = self.__validate_names_as_indices(names, self.headers)
        self.__validate_no_duplicate_names(indices)

        if set(indices) == set(self.headers.values()):
            return self.reset_matrix(None)

        indices.sort(reverse=True)
        for i in indices:
            for row in self.matrix:
                del row.values[i]

        self.reset_headers()

        return self

    def insert_rows(self, i, rows):
        if self.is_empty():
            return self.reset_matrix(rows)

        replace_headers = (i == 0)

        rows = modify_iteration_depth(iterator_to_collection(rows), depth=2)
        _, m = self.__validate_headers_and_matrix(rows, self.headers)

        if replace_headers:
            self.reset_headers(m.pop(0).values)
            i = 1
        elif m[0].is_header_row():
            del m[0]

        self.matrix[i:i] = m

        return self

    def append_rows(self, rows):
        if self.is_empty():
            return self.reset_matrix(rows)

        rows = modify_iteration_depth(iterator_to_collection(rows), depth=2)
        _, m = self.__validate_headers_and_matrix(rows, self.headers)

        if m[0].is_header_row():
            del m[0]

        self.matrix.extend(m)

        return self

    def shorten_to(self, nrows):
        if nrows == 0:
            nrows = 1
        elif nrows < 0:
            raise ValueError('nrows must be positive')

        del self.matrix[nrows + 1:]
        return self

    def join(self, other, *on_columns):
        """
        what if user wants another mapping method ?
            d = iterable.map_rows(names_b)
            d = iterable.map_rows_append(names_b)

        or different rowtype?
            d = iterable.map_rows(names_b, rowtype=namedtuple)
        """
        names = self.__validate_names_not_empty(on_columns, self.headers, depth=0)

        if isinstance(names, dict):
            names_a = list(names.keys())[0]
            names_b = list(names.values())[0]
        else:
            names_a = names
            names_b = names

        self.__validate_all_names_intersect_with_headers(names_a, self.headers)

        if isinstance(other, flux_cls):
            self.__validate_all_names_intersect_with_headers(names_b, other.headers)
            d = other.map_rows(names_b)
        elif isinstance(other, dict):
            d = other
        else:
            d = {item: item for item in other}

        rva = self.row_values_accessor(names_a)

        for row in self.matrix[1:]:
            k = rva(row)
            if k in d:
                yield row, d[k]

    def reverse(self):
        self.matrix[1:] = list(reversed(self.matrix[1:]))
        return self

    def reversed(self):
        flux = self.copy()
        flux.matrix[1:] = list(reversed(flux.matrix[1:]))
        return flux

    def sort(self, *names, reverse=False):
        """ in-place sort

        eg:
            flux.sort('col_a')
            flux.sort('col_a', 'col_b', 'col_c',
                      reverse=[True, False, True])
        """
        names = standardize_variable_arity_values(names, depth=1)
        if not names:
            return self

        self.matrix[1:] = self.__sort_rows(self.matrix[1:],
                                           names,
                                           reverse)
        return self

    def sorted(self, *names, reverse=False):
        """ :return: sorted flux_cls

        eg:
            flux = flux.sorted('col_a')
            flux = flux.sorted('col_a', 'col_b', 'col_c',
                               reverse=[False, True, True])
        """
        names = standardize_variable_arity_values(names, depth=1)
        if not names:
            return self.copy()

        flux = self.copy()
        flux.matrix[1:] = self.__sort_rows(flux.matrix[1:],
                                           names,
                                           reverse)
        return flux

    def __sort_rows(self, rows, names, reverse):
        reverse = [bool(_) for _ in standardize_variable_arity_values(reverse, depth=1)]
        reverse.extend([False] * (len(names) - len(reverse)))

        all_true  = all(_ is True  for _ in reverse)
        all_false = all(_ is False for _ in reverse)

        if all_true or all_false:
            rva = self.row_values_accessor(names)
            rows.sort(key=rva, reverse=reverse[0])

            return rows

        # last name is sorted first, first name is sorted last
        names   = reversed(names)
        reverse = reversed(reverse)

        for name, rev in zip(names, reverse):
            rva = self.row_values_accessor(name)
            rows.sort(key=rva, reverse=rev)

        return rows

    def filter(self, f, *args, **kwargs):
        """ in-place """
        self.matrix[1:] = [row for row in self.matrix[1:]
                               if f(row, *args, **kwargs)]
        return self

    def filtered(self, f, *args, **kwargs):
        """ :return: new flux_cls """
        flux = self.copy()
        flux.matrix[1:] = [row for row in flux.matrix[1:]
                               if f(row, *args, **kwargs)]
        return flux

    def filter_by_unique(self, *names):
        return self.__filter_unique_rows(*names, in_place=True)

    def filtered_by_unique(self, *names):
        return self.__filter_unique_rows(*names, in_place=False)

    def __filter_unique_rows(self, *names, in_place):
        # region {closure functions}
        u   = set()
        rva = self.row_values_accessor(names)

        def evaluate_unique(row):
            v = rva(row)

            if v not in u:
                u.add(v)
                return True
            else:
                return False
        # endregion

        if in_place:
            self.filter(evaluate_unique)
        else:
            return self.filtered(evaluate_unique)

    def map_rows(self, *names, rowtype=flux_row_cls) -> Dict[Any, flux_row_cls]:
        """ dictionary of {row_value: row} """
        items = self.__zip_row_items(names, rowtype=rowtype)
        d = ordereddict(items)

        return d

    def map_rows_append(self, *names, rowtype=flux_row_cls) -> Dict[Any, List]:
        """ dictionary of {row_value: [rows]} """
        items = self.__zip_row_items(names, rowtype=rowtype)

        d = ordereddict()
        for k, v in items:
            if k in d:  d[k].append(v)
            else:       d[k] = [v]

        return d

    def group_rows(self, *names, rowtype=flux_row_cls) -> Dict[Any, Dict]:
        """ dictionary of dictionaries {row_value_a: {row_value_b: row}} """
        d = self.map_rows(names, rowtype=rowtype)
        d = to_grouped_dict(d)

        return d

    def group_rows_append(self, *names, rowtype=flux_row_cls) -> Dict[Any, Dict]:
        """ dictionary of dictionaries {row_value_a: {row_value_b: [rows]}} """
        d = self.map_rows_append(names, rowtype=rowtype)
        d = to_grouped_dict(d)

        return d

    @deprecated('Use flux_cls.map_rows() method instead')
    def index_row(self, *names, rowtype=flux_row_cls):
        """ deprecated: Use flux_cls.map_rows() method instead
        'index' suggests method has something to do with numerical indices,
        instead of an index, as in a table of contents
        """
        return self.map_rows(names, rowtype=rowtype)

    @deprecated('Use flux_cls.map_rows_append() method instead')
    def index_rows(self, *names, rowtype=flux_row_cls):
        """ deprecated: Use flux_cls.map_rows_append() method instead
        'index' suggests method has something to do with numerical indices,
        instead of an index, as in a table of contents
        """
        return self.map_rows_append(names, rowtype=rowtype)

    def __zip_row_items(self, names, rowtype):
        rowtype = self.__validate_mapped_rowtype(rowtype)
        rva     = self.row_values_accessor(names)

        keys = (rva(row) for row in self.matrix[1:])

        if rowtype   == 'flux_row_cls': values = (row for row in self.matrix[1:])
        elif rowtype == 'dict':         values = (v for v in self.dictrows())
        elif rowtype == 'namedrow':     values = (v for v in self.namedrows())
        elif rowtype == 'namedtuple':   values = (v for v in self.namedtuples())
        elif rowtype == 'list':         values = (list(row.values)  for row in self.matrix[1:])
        elif rowtype == 'tuple':        values = (tuple(row.values) for row in self.matrix[1:])
        else:
            raise TypeError

        return zip(keys, values)

    def unique(self, *names):
        rva   = self.row_values_accessor(names)
        items = ((rva(row), None) for row in self.matrix[1:])

        return ordereddict(items).keys()

    # noinspection PyUnusedLocal
    def contiguous(self, *names):
        """ :return: yield (value, i_1, i_2) namedtuple where values are contiguous

        (contiguous values may only span a single row ie, i_1 == i_2)
        """
        contiguous_nt = namedtuple('Contiguous', ('value', 'i_1', 'i_2', 'rows'))

        i_1 = 1
        i_2 = i_1

        rva = self.row_values_accessor(names)
        v_1 = rva(self.matrix[i_1])
        v_2 = v_1

        for i_2, row in enumerate(self.matrix[i_1 + 1:], i_1 + 1):
            v_2 = rva(row)

            if v_2 != v_1:
                yield contiguous_nt(v_2, i_1, i_2 - 1, self.matrix[i_1:i_2])
                v_1 = v_2
                i_1 = i_2

        i_2 = len(self.matrix) - 1
        yield contiguous_nt(v_2, i_1, i_2, self.matrix[i_1:])

    def indices(self, start=1, step=1):
        """
        integers corresponding to each row's index position in matrix
        eg:
            [1, 2, 3, ..., len(flux.matrix)] = list(flux.indices(start=1))
            [0, 1, 2, ..., len(flux.matrix)] = list(flux.indices(start=0))

            flux['enum'] = flux.indices(start=1)
        """
        return range(start, len(self.matrix), step)

    def label_row_indices(self, start=0):
        """ meant to assist with debugging;

        label each flux_row_cls.__dict__ with an index, which will then appear
        in each row's __repr__ function and make them easier to identify after
        filtering, sorting, etc
        """
        if 'r_i' in self.headers:
            raise ColumnNameError("'r_i' already exists in headers")

        for i, row in enumerate(self.matrix, start):
            row.__dict__['r_i'] = i

        return self

    def clear_row_indices(self):
        if 'r_i' in self.headers:
            raise ColumnNameError("'r_i' already exists as header name")

        for row in self.matrix:
            if 'r_i' in row.__dict__:
                del row.__dict__['r_i']

        return self

    def copy(self, deep=False):
        if deep:
            return deepcopy(self)

        if self.__class__ is not flux_cls:
            return deepcopy(self)

        __dict__ = {k: v for k, v in self.__dict__.items()
                         if k not in {'headers', 'matrix'}}
        flux = self.__class__(self)
        flux.__dict__.update(__dict__)

        return flux

    def row_values_accessor(self, names):
        """ :return: a function that can be called for each row in self.matrix
        to retrieve column values
        """

        # region {closure functions}
        def row_value(row):
            return row.values[i]

        def row_values(row):
            return tuple([row.values[_] for _ in i])

        def row_values_slice(row):
            return tuple(row.values[i])

        def row_values_all(row):
            return tuple(row.values)
        # endregion

        i, rva = self.__validate_row_values_accessor(names, self.headers)

        rva = {'row_value':        row_value,
               'row_values':       row_values,
               'row_values_slice': row_values_slice,
               'row_values_all':   row_values_all} \
               .get(rva, rva)

        return rva

    def __getitem__(self, names):
        return self.columns(names)

    def __setitem__(self, name, values):
        """ sets values to a single column
        values are expected to exclude header row

        eg:
            flux['col'] = ['blah'] * flux.num_rows
            flux[-1]    = ['blah'] * flux.num_rows

            to insert column:
                flux[(0, 'new_col')] = ['blah'] * flux.num_rows
        """
        if isinstance(name, tuple) and len(name) == 2:
            if isinstance(name[0], int) and name[1] not in self.headers:
                i, name = name
                self.insert_columns((i, name))
            else:
                raise ColumnNameError('__setitem__ cannot be used to set values to multiple columns')

        elif isinstance(name, slice):
            raise ColumnNameError('__setitem__ cannot be used to set values to multiple columns')
        elif iteration_depth(name) > 0:
            raise ColumnNameError('__setitem__ cannot be used to set values to multiple columns')

        if name not in self.headers and not isinstance(name, int):
            self.append_columns(name, values=values)
            return

        i = self.__validate_names_as_indices(name, self.headers)[0]
        for row, v in zip(self.matrix[1:], values):
            row.values[i] = v

    def __iter__(self) -> Generator[flux_row_cls, None, None]:
        return (row for row in self.matrix[1:])

    def __add__(self, rows):
        flux = self.copy()
        flux.append_rows(rows)

        return flux

    def __iadd__(self, rows):
        return self.append_rows(rows)

    # noinspection PyTypeChecker
    def __getstate__(self):
        f = self.__class__.__init__

        constructor_params = function_parameters(f)
        constructor_repr   = '{}({})'.format(function_name(f),
                                             ', '.join(p.name for p in constructor_params))

        args   = []
        kwargs = ordereddict()
        for p in constructor_params:
            name = p.name

            if name   == 'self':      continue
            elif name == 'matrix':    p.value = [[*row.values] for row in self.matrix]
            elif name == 'headers':   p.value = ordereddict(self.headers.items())
            elif hasattr(self, name): p.value = getattr(self, name)

            elif p.value.__name__ == '_empty':
                raise ValueError("unable to resolve constructor parameter: '{}' \n"
                                 "for: {}".format(name, constructor_repr))

            kind = p.kind.lower()

            if 'positional' in kind:
                args.append(p.value)
            else:
                kwargs[name] = p.value

        args = tuple(args)

        return {'args':   args,
                'kwargs': kwargs}

    def __setstate__(self, state):
        if not isinstance(state, dict):
            return self.__init__(state)

        use_constructor = ('args'   in state and
                           'kwargs' in state)
        if use_constructor:
            self.__init__(*state['args'], **state['kwargs'])
        else:
            self.__dict__.update(state)

    def __len__(self):
        """ header row is included, see self.num_rows """
        return len(self.matrix)

    def __reversed__(self):
        return (row for row in reversed(self.matrix[1:]))

    def __repr__(self):
        if self.is_empty():
            return format_header('')

        if self.is_jagged():
            if self.has_duplicate_row_pointers(): jagged_label = '**dup**  🗲jagged🗲  '
            else:                                 jagged_label = '🗲jagged🗲  '
        else:
            jagged_label = ''

        num_rows = format_integer(len(self.matrix) - 1)
        num_rows = format_header_lite('1+' + num_rows)

        headers = ', '.join(str(n) for n in self.header_names())
        headers = format_header(headers)

        return '{}{}  {}'.format(jagged_label, num_rows, headers)

    # region {validation functions}
    @staticmethod
    def __validate_names_as_headers(names):
        headers = map_values_to_enum(names)

        conflicting = headers.keys() & set(flux_row_cls.reserved_names())
        if conflicting:
            raise ColumnNameError('column name conflict with existing headers: {}'.format(conflicting))

        return headers

    # noinspection PyProtectedMember
    @staticmethod
    def __validate_matrix(m):
        if (m is None) or (m == []):
            return [[]]

        base_cls_names = set(base_class_names(m))

        if base_cls_names & {'flux_cls', 'excel_levity_cls'}:
            return [[*row] for row in m.rows()]
        elif 'DataFrame' in base_cls_names:
            return [[*row] for row in m.values.tolist()]
        elif 'ndarray' in base_cls_names:
            return m.tolist()

        if is_exhaustable(m):
            m = list(m)
        if isinstance(m, array):
            m = list(m)
        elif isinstance(m, ItemsView):
            m = list(m)
            m = list(transpose(m, astype=list))
        elif isinstance(m, dict):
            m = list(m.items())
            m = list(transpose(m, astype=list))

        if not is_subscriptable(m):
            raise IndexError('matrix must be subscriptable')

        row_first: Union[flux_row_cls, namedtuple, dict, object]
        row:       Union[flux_row_cls, namedtuple, dict, object]

        row_first = m[0]

        # list of flux_row_cls objects
        if isinstance(row_first, flux_row_cls):
            if row_first.is_header_row():
                m = [[*row.values] for row in m]
            else:
                m = [[*row_first.header_names()]] + \
                    [[*row.values] for row in m]

        # list of dictionaries
        elif isinstance(row_first, dict):
            m = [[*row_first.keys()]] + \
                [[*row.values()] for row in m]

        # list of namedtuples
        elif hasattr(row_first, '_fields'):
            m = [[*row_first._fields]] + \
                [[*row] for row in m]

        # list of objects
        elif hasattr(row_first, '__dict__'):
            m = [[*row_first.__dict__.keys()]] + \
                [[*row.__dict__.values()] for row in m]

        # list of objects (with __slots__ attributes)
        elif hasattr(row_first, '__slots__'):
            h = tuple(row_first.__slots__)
            m = [[*h]] + \
                [[*[row.__getattribute__(n) for n in h]] for row in m]

        else:
            if iteration_depth(m, first_element_only=True) < 2:
                raise IterationDepthError('matrix must have at least two iterable dimensions (ie, a list of lists)')

        return m

    @staticmethod
    def __validate_headers_and_matrix(matrix, headers):
        matrix = flux_cls.__validate_matrix(matrix)
        if headers is None:
            headers = flux_cls.__validate_names_as_headers(matrix[0])

        return headers, [flux_row_cls(headers, row) for row in matrix]

    @staticmethod
    def __validate_row_values_accessor(names, headers):
        names = flux_cls.__validate_names_not_empty(names, depth_offset=-1)

        if callable(names):
            i = None
            f = names
        elif isinstance(names, slice):
            i = names
            if i.start in (None, 0) and \
               i.stop  in (None, len(headers)) and \
               i.step  in (None, 1):

                f = 'row_values_all'
            else:
                f = 'row_values_slice'

        elif iteration_depth(names) == 0:
            i = flux_cls.__validate_names_as_indices(names, headers)[0]
            f = 'row_value'
        else:
            i = flux_cls.__validate_names_as_indices(names, headers)
            f = 'row_values'

            if are_indices_contiguous(i):
                i = slice(i[0], i[-1] + 1)
                f = 'row_values_slice'

        return i, f

    @staticmethod
    def __validate_mapped_rowtype(rowtype):
        validtypes = {'flux_row_cls',
                      'namedrow',
                      'namedtuple',
                      'tuple',
                      'list',
                      'dict'}

        if not isinstance(rowtype, str):
            rowtype = object_name(rowtype)

        rowtype = rowtype.lower()

        if rowtype != 'flux_row_cls' and rowtype.endswith('s'):
            rowtype = rowtype[:-1]

        if rowtype not in validtypes:
            raise TypeError("invalid rowtype '{}', rowtype must be in the following: \n\t{}"
                            .format(rowtype, '\n\t'.join(validtypes)))

        return rowtype

    @staticmethod
    def __validate_names_not_empty(names,
                                   headers=None,
                                   depth=None,
                                   depth_offset=None):

        names = flux_cls.__validate_names(names, headers, depth, depth_offset)
        if is_collection(names) and len(names) == 0:
            raise ColumnNameError('no column names submitted')

        return names

    @staticmethod
    def __validate_names(names,
                         headers=None,
                         depth=None,
                         depth_offset=None):

        names = standardize_variable_arity_values(names,
                                                  depth=depth,
                                                  depth_offset=depth_offset)
        if isinstance(names, slice) and isinstance(headers, dict):
            names = list(headers.keys())[names]

        return names

    @staticmethod
    def __validate_name_datatypes(names, headers=None):
        valid_datatypes = (int, 
                           str, 
                           bytes)

        names = flux_cls.__validate_names(names, headers, depth=1)

        invalid = [n for n in names if not isinstance(n, valid_datatypes)]
        if invalid:
            s = ("invalid column name datatype: {} "
                 "\nvalid datatypes are ({})"
                 '\n\tMake sure any arguments following variable position parameters are submitted by keyword'
                 "\n\t\tflux.sort('column_a', 'column_b', reverse=True) "
                 '\n\tand not: '
                 "\n\t\tflux.sort('column_a', 'column_b', True)"
                 .format(invalid, ', '.join(v.__name__ for v in valid_datatypes)))
            raise TypeError(s)

        return names

    @staticmethod
    def __validate_inserted_items(inserted, headers):
        if isinstance(inserted, dict):
            inserted = list(inserted.items())

        inserted = standardize_variable_arity_values(inserted, depth=2)

        for item in inserted:
            if iteration_depth(item) != 1 or len(item) != 2:
                raise IterationDepthError("inserted values must be ({location}, {name}) tuples "
                                          "\n\teg: (2, 'new_col') "
                                          "\n\teg: [('col_a', 'new_col_a'), ('col_b', 'new_col_b')]")

        indices, names = list(zip(*inserted))

        flux_cls.__validate_name_datatypes(names)
        flux_cls.__validate_no_duplicate_names(names)
        flux_cls.__validate_no_names_intersect_with_headers(names, headers)
        flux_cls.__validate_all_names_intersect_with_headers(indices, headers)

        return inserted

    @staticmethod
    def __validate_no_duplicate_names(names):
        names = flux_cls.__validate_names(names, depth=1)

        duplicates = [n for n, count in Counter(names).items() if count > 1]
        if duplicates:
            raise ColumnNameError('duplicate column name detected: \n{}'.format(duplicates))

        return names

    @staticmethod
    def __validate_no_names_intersect_with_headers(names, headers):
        names = flux_cls.__validate_names(names, headers, depth=1)

        conflicting = headers.keys() & set(names)
        if conflicting:
            raise ColumnNameError('column names already exist: \n{}'.format(conflicting))

        return names

    @staticmethod
    def __validate_all_names_intersect_with_headers(names, headers):
        names = flux_cls.__validate_names(names, headers, depth=1)

        h_names = [*[ i for i in range(len(headers))],
                   *[-i for i in range(len(headers) + 1)]]
        h_names = headers.keys() | set(h_names)

        invalid = set(names) - h_names
        if invalid:
            s = '\n\t'.join((repr(n)[1:-1].replace(',', ':')
                                          .replace("'", '')
                                          .replace('"', '')) for n in headers.items())
            s = ("'{}' column name does not exist, available columns: "
                 "\n\t{}".format(invalid, s))

            raise ColumnNameError(s)

        return names

    @staticmethod
    def __validate_names_as_indices(names, headers):
        names = flux_cls.__validate_name_datatypes(names)
        names = flux_cls.__validate_all_names_intersect_with_headers(names, headers)

        h_indices = list(headers.values())
        indices = [headers.get(n, n) for n in names]       # lookup strings
        indices = [h_indices[i] for i in indices]          # lookup negative integers

        return indices

    @staticmethod
    def __validate_renamed_or_inserted_column(name, headers):
        is_remapped_column = isinstance(name, dict)
        is_inserted_column = (isinstance(name, str)   and name.startswith('(')  and name.endswith(')') or
                              isinstance(name, bytes) and name.startswith(b'(') and name.endswith(b')')
                              and name not in headers)

        if is_remapped_column:
            name_old, name = list(name.items())[0]
            headers[name]  = headers[name_old]
        elif is_inserted_column:
            name = name[1:-1]
            if name in headers:
                raise ColumnNameError("column: '{}' already exists".format(name))

        elif name not in headers:
            raise ColumnNameError("column: '{name}' does not exist. "
                                  "\nTo ensure new columns are being created intentionally (not a new column "
                                  "because to a typo) inserted headers must be surrounded by parenthesis, eg: "
                                  "\n '({name})', not '{name}'".format(name=name))

        return name

    def __validate_command_methods(self, commands, command_namedtuple):
        parsed = []
        for command in commands:
            args   = []
            kwargs = {}

            if isinstance(command, (list, tuple)):
                name = command[0]

                for a in command[1:]:
                    if isinstance(a, dict):
                        kwargs.update(a)
                    elif isinstance(a, (list, tuple)):
                        args.extend(a)
                    else:
                        args.append(a)
            else:
                name = command

            if name.startswith('__'):
                name = '_{}{}'.format(self.__class__.__name__, name)

            args   = tuple(args)
            method = getattr(self, name)

            parsed.append(command_namedtuple(name, method, args, kwargs))

        return parsed

    @staticmethod
    def __validate_profiler_function(use_profiler):
        if use_profiler in (None, False):
            return None

        if use_profiler is True:
            if line_profiler_installed:
                # noinspection PyUnresolvedReferences
                from line_profiler import LineProfiler
                return LineProfiler()
            else:
                return print_runtime

        profiler = str(use_profiler).lower()

        if profiler == 'print_runtime':
            return print_runtime

        if profiler in ('line_profiler', 'lineprofiler'):
            if line_profiler_installed is False:
                raise ImportError("'line_profiler' package not installed")

            # noinspection PyUnresolvedReferences
            from line_profiler import LineProfiler
            return LineProfiler()

        raise ValueError("invalid profiler: '{}', profiler should be in "
                         "\n(None, False, True, 'print_runtime', 'line_profiler')".format(profiler))
    # endregion




import os
import sys

from textwrap import dedent
from logging import Logger
from logging import Formatter
from logging import FileHandler
from logging import StreamHandler
from logging import (NOTSET,
                     DEBUG,
                     INFO,
                     WARNING,
                     ERROR,
                     CRITICAL)

from .. util.filesystem import parse_path
from .. util.filesystem import parse_file_name
from .. util.filesystem import standardize_dir
from .. util.text import object_name
from .. util.text import styled

from .. conditional import is_utf_console
from .. conditional import is_tty_console


class log_cls(Logger):
    banner_character = '*'
    banner_width     = None

    def __init__(self, path_or_name='',
                       level='DEBUG',
                       log_format='%(message)s',
                       exception_callback=None,
                       colored_statements=False):
        """
        :param path_or_name:
            parsed to determine name of logger
            if path_or_name includes directory or file extension, a file handler is added
        :param log_format:
            format to be applied to handlers
            eg log_format:
                '[%(asctime)s] [%(levelname)s] %(message)s'
        :param exception_callback:
            function to be invoked when self.exception_handler() is called
            sys.excepthook is automatically set to self.exception_handler if
            exception_callback is a valid function
        """
        if (exception_callback is not None) and not callable(exception_callback):
            raise TypeError('exception_callback must be callable')

        name  = parse_file_name(path_or_name)
        level = level.upper()

        super().__init__(name, level)

        self.path               = self._set_path(path_or_name)
        self.formatter          = Formatter(log_format)
        self.exception_callback = exception_callback
        self.exception_message  = ''

        self._add_stream_handler(sys.stdout, colored_statements)
        self._add_file_handler(self.path)

        if exception_callback:
            sys.excepthook = self.exception_handler

    @property
    def stream_handlers(self):
        handlers = []
        for h in self.handlers:
            if type(h) in (StreamHandler, colored_streamhandler_cls):
                handlers.append(h)

        return handlers

    @property
    def file_handlers(self):
        handlers = []
        for h in self.handlers:
            if isinstance(h, FileHandler):
                handlers.append(h)

        return handlers

    @property
    def log_format(self):
        # noinspection PyProtectedMember
        return self.formatter._fmt

    def set_format(self, log_format):
        self.formatter = Formatter(log_format)
        for h in self.handlers:
            h.setFormatter(self.formatter)

    def unformat(self, message):
        # noinspection PyProtectedMember
        lf = (self.formatter._fmt
                  .replace('%(asctime)s', '')
                  .replace('%(levelname)s', '')
                  .replace('%(message)s', ''))
        return message.replace(lf, '')

    def add_parent_log(self, p_log):
        if id(p_log) == id(self):
            raise ValueError('parent log and self are the same')

        self.parent = p_log
        self.close_stream_handlers()

    def close(self):
        self.close_stream_handlers()
        self.close_file_handlers()

    def close_stream_handlers(self):
        for h in self.stream_handlers:
            h.close()
            self.removeHandler(h)

    def close_file_handlers(self):
        for h in self.file_handlers:
            h.close()
            self.removeHandler(h)

    def _add_stream_handler(self, stream, colored_statements):
        if not colored_statements:
            h = StreamHandler(stream)
        elif not is_utf_console:
            h = StreamHandler(stream)
        elif is_tty_console:
            h = StreamHandler(stream)
        else:
            h = colored_streamhandler_cls(stream)

        h.setLevel(self.level)
        h.setFormatter(self.formatter)
        self.addHandler(h)

    def _add_file_handler(self, path):
        if not path:
            return

        h = FileHandler(path, mode='w', encoding='utf-8')

        h.setLevel(self.level)
        h.setFormatter(self.formatter)
        self.addHandler(h)

    def exception_handler(self, e_type, e_msg, e_traceback):
        """
        sys.excepthook = log.exception_handler

        try:
            1 / 0
        except:
            log.exception_handler(*sys.exc_info())
        """

        self.exception_message = self._formatted_exception_message(e_type, e_msg, e_traceback)
        self.error(self.exception_message, exc_info=(e_type, e_msg, e_traceback))

        if self.exception_callback:
            self.exception_callback()

        return self.exception_message

    def _formatted_exception_message(self, e_type, e_msg, e_traceback):
        title_message = '(The result W:Resign was added to the game information)'

        _e_type_ = 'Exception'
        _e_msg_  = 'unknown error'
        filename = 'unknown'
        lineno   = 'unknown'

        if e_type:
            _e_type_ = object_name(e_type)

        if e_msg:
            _e_msg_ = '{}'.format(str(e_msg).replace('"', "'"))

        if e_traceback:
            filename = e_traceback.tb_frame.f_code.co_filename
            lineno   = e_traceback.tb_lineno

        if isinstance(self.banner_width, int):
            banner_width = self.banner_width
        else:
            banner_width = max(len(title_message),
                               *[len(line) for line in str(_e_msg_).split('\n')],
                               *[len(line) for line in str(filename).split('\n')],
                               80) + 10

        if isinstance(self.banner_character, str):
            _bf_ = '{:%s^%i}' % (self.banner_character, banner_width)
            banner_upper = _bf_.format(' __{}__ '.format(self.name))
            banner_lower = self.banner_character * banner_width
        else:
            banner_upper = ''
            banner_lower = ''

        exception_message = '''
        {banner_upper}
            {title_message}

            <{e_type}> {e_msg}
            File: {filename}
            Line: {lineno}
            {repr_self}
        {banner_lower}
        '''.format(banner_upper=banner_upper,
                   title_message=title_message,
                   e_type=_e_type_,
                   e_msg=_e_msg_,
                   filename=filename,
                   lineno=lineno,
                   repr_self=repr(self),
                   banner_lower=banner_lower)

        exception_message = dedent(exception_message)

        return exception_message

    @staticmethod
    def _set_path(path):
        p_path = parse_path(path, explicit_cwd=False)

        directory = p_path.directory
        filename  = p_path.filename
        extension = p_path.extension

        if not (directory or extension):
            return ''

        directory = standardize_dir(directory, explicit_cwd=True)
        if not os.path.exists(directory):
            raise FileExistsError('log directory does not exist: \n{}'.format(directory))
            # os.makedirs(directory)

        if extension == '.py':
            extension = '.log'
        elif extension == '':
            extension = '.log'

        path = directory + filename + extension

        return path

    def __repr__(self):
        return 'vengeance log: {}'.format(self.name)


class colored_streamhandler_cls(StreamHandler):
    level_colors  = {NOTSET:   'grey',
                     DEBUG:    'grey',
                     INFO:     'white',
                     WARNING:  'yellow',
                     ERROR:    'bright red',
                     CRITICAL: 'bright magenta'}

    def __init__(self, stream=None):
        super().__init__(stream)

    # noinspection PyBroadException
    def emit(self, record):
        try:
            color = self.level_colors.get(record.levelno, 'grey')
            if record.levelno == CRITICAL:
                effect = 'bold|underline'
            else:
                effect = 'bold'

            s = self.format(record) + self.terminator
            s = styled(s, color, effect)

            self.flush()
            self.stream.write(s)

        except RecursionError:
            raise
        except Exception:
            self.handleError(record)





# Excel's integer values for cell errors
excel_errors = {-2146826281: '#excel_error: div0#',
                -2146826246: '#excel_error: na#',
                -2146826259: '#excel_error: name#',
                -2146826288: '#excel_error: null#',
                -2146826252: '#excel_error: num#',
                -2146826265: '#excel_error: ref#',
                -2146826273: '#excel_error: value#'}

# region {excel constants}
# site-packages\win32com\gen_py\00020813-0000-0000-C000-000000000046x0x1x9\__init__.py

xl3DBar                                   = -4099
xl3DEffects1                              = 13
xl3DEffects2                              = 14
xl3DSurface                               = -4103
xlAbove                                   = 0
xlAccounting1                             = 4
xlAccounting2                             = 5
xlAccounting3                             = 6
xlAccounting4                             = 17
xlAdd                                     = 2
xlAll                                     = -4104
xlAllExceptBorders                        = 7
xlAutomatic                               = -4105
xlBar                                     = 2
xlBelow                                   = 1
xlBidi                                    = -5000
xlBidiCalendar                            = 3
xlBoth                                    = 1
xlBottom                                  = -4107
xlCascade                                 = 7
xlCenter                                  = -4108
xlCenterAcrossSelection                   = 7
xlChart4                                  = 2
xlChartSeries                             = 17
xlChartShort                              = 6
xlChartTitles                             = 18
xlChecker                                 = 9
xlCircle                                  = 8
xlClassic1                                = 1
xlClassic2                                = 2
xlClassic3                                = 3
xlClosed                                  = 3
xlColor1                                  = 7
xlColor2                                  = 8
xlColor3                                  = 9
xlColumn                                  = 3
xlCombination                             = -4111
xlComplete                                = 4
xlConstants                               = 2
xlContents                                = 2
xlContext                                 = -5002
xlCorner                                  = 2
xlCrissCross                              = 16
xlCross                                   = 4
xlCustom                                  = -4114
xlDebugCodePane                           = 13
xlDefaultAutoFormat                       = -1
xlDesktop                                 = 9
xlDiamond                                 = 2
xlDirect                                  = 1
xlDistributed                             = -4117
xlDivide                                  = 5
xlDoubleAccounting                        = 5
xlDoubleClosed                            = 5
xlDoubleOpen                              = 4
xlDoubleQuote                             = 1
xlDrawingObject                           = 14
xlEntireChart                             = 20
xlExcelMenus                              = 1
xlExtended                                = 3
xlFill                                    = 5
xlFirst                                   = 0
xlFixedValue                              = 1
xlFloating                                = 5
xlFormats                                 = -4122
xlFormula                                 = 5
xlFullScript                              = 1
xlGeneral                                 = 1
xlGray16                                  = 17
xlGray25                                  = -4124
xlGray50                                  = -4125
xlGray75                                  = -4126
xlGray8                                   = 18
xlGregorian                               = 2
xlGrid                                    = 15
xlGridline                                = 22
xlHigh                                    = -4127
xlHindiNumerals                           = 3
xlIcons                                   = 1
xlImmediatePane                           = 12
xlInside                                  = 2
xlInteger                                 = 2
xlJustify                                 = -4130
xlLTR                                     = -5003
xlLast                                    = 1
xlLastCell                                = 11
xlLatin                                   = -5001
xlLeft                                    = -4131
xlLeftToRight                             = 2
xlLightDown                               = 13
xlLightHorizontal                         = 11
xlLightUp                                 = 14
xlLightVertical                           = 12
xlList1                                   = 10
xlList2                                   = 11
xlList3                                   = 12
xlLocalFormat1                            = 15
xlLocalFormat2                            = 16
xlLogicalCursor                           = 1
xlLong                                    = 3
xlLotusHelp                               = 2
xlLow                                     = -4134
xlMacrosheetCell                          = 7
xlManual                                  = -4135
xlMaximum                                 = 2
xlMinimum                                 = 4
xlMinusValues                             = 3
xlMixed                                   = 2
xlMixedAuthorizedScript                   = 4
xlMixedScript                             = 3
xlModule                                  = -4141
xlMultiply                                = 4
xlNarrow                                  = 1
xlNextToAxis                              = 4
xlNoDocuments                             = 3
xlNone                                    = -4142
xlNotes                                   = -4144
xlOff                                     = -4146
xlOn                                      = 1
xlOpaque                                  = 3
xlOpen                                    = 2
xlOutside                                 = 3
xlPartial                                 = 3
xlPartialScript                           = 2
xlPercent                                 = 2
xlPlus                                    = 9
xlPlusValues                              = 2
xlRTL                                     = -5004
xlReference                               = 4
xlRight                                   = -4152
xlScale                                   = 3
xlSemiGray75                              = 10
xlSemiautomatic                           = 2
xlShort                                   = 1
xlShowLabel                               = 4
xlShowLabelAndPercent                     = 5
xlShowPercent                             = 3
xlShowValue                               = 2
xlSimple                                  = -4154
xlSingle                                  = 2
xlSingleAccounting                        = 4
xlSingleQuote                             = 2
xlSolid                                   = 1
xlSquare                                  = 1
xlStError                                 = 4
xlStar                                    = 5
xlStrict                                  = 2
xlSubtract                                = 3
xlSystem                                  = 1
xlTextBox                                 = 16
xlTiled                                   = 1
xlTitleBar                                = 8
xlToolbar                                 = 1
xlToolbarButton                           = 2
xlTop                                     = -4160
xlTopToBottom                             = 1
xlTransparent                             = 2
xlTriangle                                = 3
xlVeryHidden                              = 2
xlVisible                                 = 12
xlVisualCursor                            = 2
xlWatchPane                               = 11
xlWide                                    = 3
xlWorkbookTab                             = 6
xlWorksheet4                              = 1
xlWorksheetCell                           = 3
xlWorksheetShort                          = 5
xlAboveAverage                            = 0
xlAboveStdDev                             = 4
xlBelowAverage                            = 1
xlBelowStdDev                             = 5
xlEqualAboveAverage                       = 2
xlEqualBelowAverage                       = 3
xlActionTypeDrillthrough                  = 256
xlActionTypeReport                        = 128
xlActionTypeRowset                        = 16
xlActionTypeUrl                           = 1
xlAutomaticAllocation                     = 2
xlManualAllocation                        = 1
xlEqualAllocation                         = 1
xlWeightedAllocation                      = 2
xlAllocateIncrement                       = 2
xlAllocateValue                           = 1
xl24HourClock                             = 33
xl4DigitYears                             = 43
xlAlternateArraySeparator                 = 16
xlColumnSeparator                         = 14
xlCountryCode                             = 1
xlCountrySetting                          = 2
xlCurrencyBefore                          = 37
xlCurrencyCode                            = 25
xlCurrencyDigits                          = 27
xlCurrencyLeadingZeros                    = 40
xlCurrencyMinusSign                       = 38
xlCurrencyNegative                        = 28
xlCurrencySpaceBefore                     = 36
xlCurrencyTrailingZeros                   = 39
xlDateOrder                               = 32
xlDateSeparator                           = 17
xlDayCode                                 = 21
xlDayLeadingZero                          = 42
xlDecimalSeparator                        = 3
xlGeneralFormatName                       = 26
xlHourCode                                = 22
xlLeftBrace                               = 12
xlLeftBracket                             = 10
xlListSeparator                           = 5
xlLowerCaseColumnLetter                   = 9
xlLowerCaseRowLetter                      = 8
xlMDY                                     = 44
xlMetric                                  = 35
xlMinuteCode                              = 23
xlMonthCode                               = 20
xlMonthLeadingZero                        = 41
xlMonthNameChars                          = 30
xlNonEnglishFunctions                     = 34
xlNoncurrencyDigits                       = 29
xlRightBrace                              = 13
xlRightBracket                            = 11
xlRowSeparator                            = 15
xlSecondCode                              = 24
xlThousandsSeparator                      = 4
xlTimeLeadingZero                         = 45
xlTimeSeparator                           = 18
xlUICultureTag                            = 46
xlUpperCaseColumnLetter                   = 7
xlUpperCaseRowLetter                      = 6
xlWeekdayNameChars                        = 31
xlYearCode                                = 19
xlColumnThenRow                           = 2
xlRowThenColumn                           = 1
xlArabicBothStrict                        = 3
xlArabicNone                              = 0
xlArabicStrictAlefHamza                   = 1
xlArabicStrictFinalYaa                    = 2
xlArrangeStyleCascade                     = 7
xlArrangeStyleHorizontal                  = -4128
xlArrangeStyleTiled                       = 1
xlArrangeStyleVertical                    = -4166
xlArrowHeadLengthLong                     = 3
xlArrowHeadLengthMedium                   = -4138
xlArrowHeadLengthShort                    = 1
xlArrowHeadStyleClosed                    = 3
xlArrowHeadStyleDoubleClosed              = 5
xlArrowHeadStyleDoubleOpen                = 4
xlArrowHeadStyleNone                      = -4142
xlArrowHeadStyleOpen                      = 2
xlArrowHeadWidthMedium                    = -4138
xlArrowHeadWidthNarrow                    = 1
xlArrowHeadWidthWide                      = 3
xlFillCopy                                = 1
xlFillDays                                = 5
xlFillDefault                             = 0
xlFillFormats                             = 3
xlFillMonths                              = 7
xlFillSeries                              = 2
xlFillValues                              = 4
xlFillWeekdays                            = 6
xlFillYears                               = 8
xlFlashFill                               = 11
xlGrowthTrend                             = 10
xlLinearTrend                             = 9
xlAnd                                     = 1
xlBottom10Items                           = 4
xlBottom10Percent                         = 6
xlFilterAutomaticFontColor                = 13
xlFilterCellColor                         = 8
xlFilterDynamic                           = 11
xlFilterFontColor                         = 9
xlFilterIcon                              = 10
xlFilterNoFill                            = 12
xlFilterNoIcon                            = 14
xlFilterValues                            = 7
xlOr                                      = 2
xlTop10Items                              = 3
xlTop10Percent                            = 5
xlAxisCrossesAutomatic                    = -4105
xlAxisCrossesCustom                       = -4114
xlAxisCrossesMaximum                      = 2
xlAxisCrossesMinimum                      = 4
xlPrimary                                 = 1
xlSecondary                               = 2
xlCategory                                = 1
xlSeriesAxis                              = 3
xlValue                                   = 2
xlBackgroundAutomatic                     = -4105
xlBackgroundOpaque                        = 3
xlBackgroundTransparent                   = 2
xlBox                                     = 0
xlConeToMax                               = 5
xlConeToPoint                             = 4
xlCylinder                                = 3
xlPyramidToMax                            = 2
xlPyramidToPoint                          = 1
xlBinsTypeAutomatic                       = 0
xlBinsTypeBinCount                        = 4
xlBinsTypeBinSize                         = 3
xlBinsTypeCategorical                     = 1
xlBinsTypeManual                          = 2
xlHairline                                = 1
xlMedium                                  = -4138
xlThick                                   = 4
xlThin                                    = 2
xlDiagonalDown                            = 5
xlDiagonalUp                              = 6
xlEdgeBottom                              = 9
xlEdgeLeft                                = 7
xlEdgeRight                               = 10
xlEdgeTop                                 = 8
xlInsideHorizontal                        = 12
xlInsideVertical                          = 11
_xlDialogChartSourceData                  = 541
_xlDialogPhonetic                         = 538
xlDialogActivate                          = 103
xlDialogActiveCellFont                    = 476
xlDialogAddChartAutoformat                = 390
xlDialogAddinManager                      = 321
xlDialogAlignment                         = 43
xlDialogAppMove                           = 170
xlDialogAppSize                           = 171
xlDialogApplyNames                        = 133
xlDialogApplyStyle                        = 212
xlDialogArrangeAll                        = 12
xlDialogAssignToObject                    = 213
xlDialogAssignToTool                      = 293
xlDialogAttachText                        = 80
xlDialogAttachToolbars                    = 323
xlDialogAutoCorrect                       = 485
xlDialogAxes                              = 78
xlDialogBorder                            = 45
xlDialogCalculation                       = 32
xlDialogCellProtection                    = 46
xlDialogChangeLink                        = 166
xlDialogChartAddData                      = 392
xlDialogChartLocation                     = 527
xlDialogChartOptionsDataLabelMultiple     = 724
xlDialogChartOptionsDataLabels            = 505
xlDialogChartOptionsDataTable             = 506
xlDialogChartSourceData                   = 540
xlDialogChartTrend                        = 350
xlDialogChartType                         = 526
xlDialogChartWizard                       = 288
xlDialogCheckboxProperties                = 435
xlDialogClear                             = 52
xlDialogColorPalette                      = 161
xlDialogColumnWidth                       = 47
xlDialogCombination                       = 73
xlDialogConditionalFormatting             = 583
xlDialogConsolidate                       = 191
xlDialogCopyChart                         = 147
xlDialogCopyPicture                       = 108
xlDialogCreateList                        = 796
xlDialogCreateNames                       = 62
xlDialogCreatePublisher                   = 217
xlDialogCreateRelationship                = 1272
xlDialogCustomViews                       = 493
xlDialogCustomizeToolbar                  = 276
xlDialogDataDelete                        = 36
xlDialogDataLabel                         = 379
xlDialogDataLabelMultiple                 = 723
xlDialogDataSeries                        = 40
xlDialogDataValidation                    = 525
xlDialogDefineName                        = 61
xlDialogDefineStyle                       = 229
xlDialogDeleteFormat                      = 111
xlDialogDeleteName                        = 110
xlDialogDemote                            = 203
xlDialogDisplay                           = 27
xlDialogDocumentInspector                 = 862
xlDialogEditColor                         = 223
xlDialogEditDelete                        = 54
xlDialogEditSeries                        = 228
xlDialogEditboxProperties                 = 438
xlDialogEditionOptions                    = 251
xlDialogErrorChecking                     = 732
xlDialogErrorbarX                         = 463
xlDialogErrorbarY                         = 464
xlDialogEvaluateFormula                   = 709
xlDialogExternalDataProperties            = 530
xlDialogExtract                           = 35
xlDialogFileDelete                        = 6
xlDialogFileSharing                       = 481
xlDialogFillGroup                         = 200
xlDialogFillWorkgroup                     = 301
xlDialogFilter                            = 447
xlDialogFilterAdvanced                    = 370
xlDialogFindFile                          = 475
xlDialogFont                              = 26
xlDialogFontProperties                    = 381
xlDialogForecastETS                       = 1300
xlDialogFormatAuto                        = 269
xlDialogFormatChart                       = 465
xlDialogFormatCharttype                   = 423
xlDialogFormatFont                        = 150
xlDialogFormatLegend                      = 88
xlDialogFormatMain                        = 225
xlDialogFormatMove                        = 128
xlDialogFormatNumber                      = 42
xlDialogFormatOverlay                     = 226
xlDialogFormatSize                        = 129
xlDialogFormatText                        = 89
xlDialogFormulaFind                       = 64
xlDialogFormulaGoto                       = 63
xlDialogFormulaReplace                    = 130
xlDialogFunctionWizard                    = 450
xlDialogGallery3dArea                     = 193
xlDialogGallery3dBar                      = 272
xlDialogGallery3dColumn                   = 194
xlDialogGallery3dLine                     = 195
xlDialogGallery3dPie                      = 196
xlDialogGallery3dSurface                  = 273
xlDialogGalleryArea                       = 67
xlDialogGalleryBar                        = 68
xlDialogGalleryColumn                     = 69
xlDialogGalleryCustom                     = 388
xlDialogGalleryDoughnut                   = 344
xlDialogGalleryLine                       = 70
xlDialogGalleryPie                        = 71
xlDialogGalleryRadar                      = 249
xlDialogGalleryScatter                    = 72
xlDialogGoalSeek                          = 198
xlDialogGridlines                         = 76
xlDialogImportTextFile                    = 666
xlDialogInsert                            = 55
xlDialogInsertHyperlink                   = 596
xlDialogInsertNameLabel                   = 496
xlDialogInsertObject                      = 259
xlDialogInsertPicture                     = 342
xlDialogInsertTitle                       = 380
xlDialogLabelProperties                   = 436
xlDialogListboxProperties                 = 437
xlDialogMacroOptions                      = 382
xlDialogMailEditMailer                    = 470
xlDialogMailLogon                         = 339
xlDialogMailNextLetter                    = 378
xlDialogMainChart                         = 85
xlDialogMainChartType                     = 185
xlDialogManageRelationships               = 1271
xlDialogMenuEditor                        = 322
xlDialogMove                              = 262
xlDialogMyPermission                      = 834
xlDialogNameManager                       = 977
xlDialogNew                               = 119
xlDialogNewName                           = 978
xlDialogNewWebQuery                       = 667
xlDialogNote                              = 154
xlDialogObjectProperties                  = 207
xlDialogObjectProtection                  = 214
xlDialogOpen                              = 1
xlDialogOpenLinks                         = 2
xlDialogOpenMail                          = 188
xlDialogOpenText                          = 441
xlDialogOptionsCalculation                = 318
xlDialogOptionsChart                      = 325
xlDialogOptionsEdit                       = 319
xlDialogOptionsGeneral                    = 356
xlDialogOptionsListsAdd                   = 458
xlDialogOptionsME                         = 647
xlDialogOptionsTransition                 = 355
xlDialogOptionsView                       = 320
xlDialogOutline                           = 142
xlDialogOverlay                           = 86
xlDialogOverlayChartType                  = 186
xlDialogPageSetup                         = 7
xlDialogParse                             = 91
xlDialogPasteNames                        = 58
xlDialogPasteSpecial                      = 53
xlDialogPatterns                          = 84
xlDialogPermission                        = 832
xlDialogPhonetic                          = 656
xlDialogPivotCalculatedField              = 570
xlDialogPivotCalculatedItem               = 572
xlDialogPivotClientServerSet              = 689
xlDialogPivotDefaultLayout                = 1360
xlDialogPivotFieldGroup                   = 433
xlDialogPivotFieldProperties              = 313
xlDialogPivotFieldUngroup                 = 434
xlDialogPivotShowPages                    = 421
xlDialogPivotSolveOrder                   = 568
xlDialogPivotTableOptions                 = 567
xlDialogPivotTableSlicerConnections       = 1183
xlDialogPivotTableWhatIfAnalysisSettings  = 1153
xlDialogPivotTableWizard                  = 312
xlDialogPlacement                         = 300
xlDialogPrint                             = 8
xlDialogPrintPreview                      = 222
xlDialogPrinterSetup                      = 9
xlDialogPromote                           = 202
xlDialogProperties                        = 474
xlDialogPropertyFields                    = 754
xlDialogProtectDocument                   = 28
xlDialogProtectSharing                    = 620
xlDialogPublishAsWebPage                  = 653
xlDialogPushbuttonProperties              = 445
xlDialogRecommendedPivotTables            = 1258
xlDialogReplaceFont                       = 134
xlDialogRoutingSlip                       = 336
xlDialogRowHeight                         = 127
xlDialogRun                               = 17
xlDialogSaveAs                            = 5
xlDialogSaveCopyAs                        = 456
xlDialogSaveNewObject                     = 208
xlDialogSaveWorkbook                      = 145
xlDialogSaveWorkspace                     = 285
xlDialogScale                             = 87
xlDialogScenarioAdd                       = 307
xlDialogScenarioCells                     = 305
xlDialogScenarioEdit                      = 308
xlDialogScenarioMerge                     = 473
xlDialogScenarioSummary                   = 311
xlDialogScrollbarProperties               = 420
xlDialogSearch                            = 731
xlDialogSelectSpecial                     = 132
xlDialogSendMail                          = 189
xlDialogSeriesAxes                        = 460
xlDialogSeriesOptions                     = 557
xlDialogSeriesOrder                       = 466
xlDialogSeriesShape                       = 504
xlDialogSeriesX                           = 461
xlDialogSeriesY                           = 462
xlDialogSetBackgroundPicture              = 509
xlDialogSetMDXEditor                      = 1208
xlDialogSetManager                        = 1109
xlDialogSetPrintTitles                    = 23
xlDialogSetTupleEditorOnColumns           = 1108
xlDialogSetTupleEditorOnRows              = 1107
xlDialogSetUpdateStatus                   = 159
xlDialogShowDetail                        = 204
xlDialogShowToolbar                       = 220
xlDialogSize                              = 261
xlDialogSlicerCreation                    = 1182
xlDialogSlicerPivotTableConnections       = 1184
xlDialogSlicerSettings                    = 1179
xlDialogSort                              = 39
xlDialogSortSpecial                       = 192
xlDialogSparklineInsertColumn             = 1134
xlDialogSparklineInsertLine               = 1133
xlDialogSparklineInsertWinLoss            = 1135
xlDialogSplit                             = 137
xlDialogStandardFont                      = 190
xlDialogStandardWidth                     = 472
xlDialogStyle                             = 44
xlDialogSubscribeTo                       = 218
xlDialogSubtotalCreate                    = 398
xlDialogSummaryInfo                       = 474
xlDialogTabOrder                          = 394
xlDialogTable                             = 41
xlDialogTextToColumns                     = 422
xlDialogUnhide                            = 94
xlDialogUpdateLink                        = 201
xlDialogVbaInsertFile                     = 328
xlDialogVbaMakeAddin                      = 478
xlDialogVbaProcedureDefinition            = 330
xlDialogView3d                            = 197
xlDialogWebOptionsBrowsers                = 773
xlDialogWebOptionsEncoding                = 686
xlDialogWebOptionsFiles                   = 684
xlDialogWebOptionsFonts                   = 687
xlDialogWebOptionsGeneral                 = 683
xlDialogWebOptionsPictures                = 685
xlDialogWindowMove                        = 14
xlDialogWindowSize                        = 13
xlDialogWorkbookAdd                       = 281
xlDialogWorkbookCopy                      = 283
xlDialogWorkbookInsert                    = 354
xlDialogWorkbookMove                      = 282
xlDialogWorkbookName                      = 386
xlDialogWorkbookNew                       = 302
xlDialogWorkbookOptions                   = 284
xlDialogWorkbookProtect                   = 417
xlDialogWorkbookTabSplit                  = 415
xlDialogWorkbookUnhide                    = 384
xlDialogWorkgroup                         = 199
xlDialogWorkspace                         = 95
xlDialogZoom                              = 256
xlErrBlocked                              = 2047
xlErrCalc                                 = 2050
xlErrConnect                              = 2046
xlErrDiv0                                 = 2007
xlErrField                                = 2049
xlErrGettingData                          = 2043
xlErrNA                                   = 2042
xlErrName                                 = 2029
xlErrNull                                 = 2000
xlErrNum                                  = 2036
xlErrRef                                  = 2023
xlErrSpill                                = 2045
xlErrUnknown                              = 2048
xlErrValue                                = 2015
xlAllValues                               = 0
xlColGroups                               = 2
xlRowGroups                               = 1
xlNumberFormatTypeDefault                 = 0
xlNumberFormatTypeNumber                  = 1
xlNumberFormatTypePercent                 = 2
xlCalculatedMeasure                       = 2
xlCalculatedMember                        = 0
xlCalculatedSet                           = 1
xlCalculationAutomatic                    = -4105
xlCalculationManual                       = -4135
xlCalculationSemiautomatic                = 2
xlAnyKey                                  = 2
xlEscKey                                  = 1
xlNoKey                                   = 0
xlCalculating                             = 1
xlDone                                    = 0
xlPending                                 = 2
xlCategoryLabelLevelAll                   = -1
xlCategoryLabelLevelCustom                = -2
xlCategoryLabelLevelNone                  = -3
xlCategoryAscending                       = 2
xlCategoryDescending                      = 3
xlIndexAscending                          = 0
xlIndexDescending                         = 1
xlAutomaticScale                          = -4105
xlCategoryScale                           = 2
xlTimeScale                               = 3
xlCellChangeApplied                       = 3
xlCellChanged                             = 2
xlCellNotChanged                          = 1
xlInsertDeleteCells                       = 1
xlInsertEntireRows                        = 2
xlOverwriteCells                          = 0
xlCellTypeAllFormatConditions             = -4172
xlCellTypeAllValidation                   = -4174
xlCellTypeBlanks                          = 4
xlCellTypeComments                        = -4144
xlCellTypeConstants                       = 2
xlCellTypeFormulas                        = -4123
xlCellTypeLastCell                        = 11
xlCellTypeSameFormatConditions            = -4173
xlCellTypeSameValidation                  = -4175
xlCellTypeVisible                         = 12
xlChartElementPositionAutomatic           = -4105
xlChartElementPositionCustom              = -4114
xlAnyGallery                              = 23
xlBuiltIn                                 = 21
xlUserDefined                             = 22
xlAxis                                    = 21
xlAxisTitle                               = 17
xlChartArea                               = 2
xlChartTitle                              = 4
xlCorners                                 = 6
xlDataLabel                               = 0
xlDataTable                               = 7
xlDisplayUnitLabel                        = 30
xlDownBars                                = 20
xlDropLines                               = 26
xlErrorBars                               = 9
xlFloor                                   = 23
xlHiLoLines                               = 25
xlLeaderLines                             = 29
xlLegend                                  = 24
xlLegendEntry                             = 12
xlLegendKey                               = 13
xlMajorGridlines                          = 15
xlMinorGridlines                          = 16
xlNothing                                 = 28
xlPivotChartCollapseEntireFieldButton     = 34
xlPivotChartDropZone                      = 32
xlPivotChartExpandEntireFieldButton       = 33
xlPivotChartFieldButton                   = 31
xlPlotArea                                = 19
xlRadarAxisLabels                         = 27
xlSeries                                  = 3
xlSeriesLines                             = 22
xlShape                                   = 14
xlTrendline                               = 8
xlUpBars                                  = 18
xlWalls                                   = 5
xlXErrorBars                              = 10
xlYErrorBars                              = 11
xlLocationAsNewSheet                      = 1
xlLocationAsObject                        = 2
xlLocationAutomatic                       = 3
xlAllFaces                                = 7
xlEnd                                     = 2
xlEndSides                                = 3
xlFront                                   = 4
xlFrontEnd                                = 6
xlFrontSides                              = 5
xlSides                                   = 1
xlStack                                   = 2
xlStackScale                              = 3
xlStretch                                 = 1
xlSplitByCustomSplit                      = 4
xlSplitByPercentValue                     = 3
xlSplitByPosition                         = 1
xlSplitByValue                            = 2
xl3DArea                                  = -4098
xl3DAreaStacked                           = 78
xl3DAreaStacked100                        = 79
xl3DBarClustered                          = 60
xl3DBarStacked                            = 61
xl3DBarStacked100                         = 62
xl3DColumn                                = -4100
xl3DColumnClustered                       = 54
xl3DColumnStacked                         = 55
xl3DColumnStacked100                      = 56
xl3DLine                                  = -4101
xl3DPie                                   = -4102
xl3DPieExploded                           = 70
xlArea                                    = 1
xlAreaStacked                             = 76
xlAreaStacked100                          = 77
xlBarClustered                            = 57
xlBarOfPie                                = 71
xlBarStacked                              = 58
xlBarStacked100                           = 59
xlBoxwhisker                              = 121
xlBubble                                  = 15
xlBubble3DEffect                          = 87
xlColumnClustered                         = 51
xlColumnStacked                           = 52
xlColumnStacked100                        = 53
xlConeBarClustered                        = 102
xlConeBarStacked                          = 103
xlConeBarStacked100                       = 104
xlConeCol                                 = 105
xlConeColClustered                        = 99
xlConeColStacked                          = 100
xlConeColStacked100                       = 101
xlCylinderBarClustered                    = 95
xlCylinderBarStacked                      = 96
xlCylinderBarStacked100                   = 97
xlCylinderCol                             = 98
xlCylinderColClustered                    = 92
xlCylinderColStacked                      = 93
xlCylinderColStacked100                   = 94
xlDoughnut                                = -4120
xlDoughnutExploded                        = 80
xlFunnel                                  = 123
xlHistogram                               = 118
xlLine                                    = 4
xlLineMarkers                             = 65
xlLineMarkersStacked                      = 66
xlLineMarkersStacked100                   = 67
xlLineStacked                             = 63
xlLineStacked100                          = 64
xlPareto                                  = 122
xlPie                                     = 5
xlPieExploded                             = 69
xlPieOfPie                                = 68
xlPyramidBarClustered                     = 109
xlPyramidBarStacked                       = 110
xlPyramidBarStacked100                    = 111
xlPyramidCol                              = 112
xlPyramidColClustered                     = 106
xlPyramidColStacked                       = 107
xlPyramidColStacked100                    = 108
xlRadar                                   = -4151
xlRadarFilled                             = 82
xlRadarMarkers                            = 81
xlRegionMap                               = 140
xlStockHLC                                = 88
xlStockOHLC                               = 89
xlStockVHLC                               = 90
xlStockVOHLC                              = 91
xlSunburst                                = 120
xlSurface                                 = 83
xlSurfaceTopView                          = 85
xlSurfaceTopViewWireframe                 = 86
xlSurfaceWireframe                        = 84
xlTreemap                                 = 117
xlWaterfall                               = 119
xlXYScatter                               = -4169
xlXYScatterLines                          = 74
xlXYScatterLinesNoMarkers                 = 75
xlXYScatterSmooth                         = 72
xlXYScatterSmoothNoMarkers                = 73
xlCheckInMajorVersion                     = 1
xlCheckInMinorVersion                     = 0
xlCheckInOverwriteVersion                 = 2
xlClipboardFormatBIFF                     = 8
xlClipboardFormatBIFF12                   = 63
xlClipboardFormatBIFF2                    = 18
xlClipboardFormatBIFF3                    = 20
xlClipboardFormatBIFF4                    = 30
xlClipboardFormatBinary                   = 15
xlClipboardFormatBitmap                   = 9
xlClipboardFormatCGM                      = 13
xlClipboardFormatCSV                      = 5
xlClipboardFormatDIF                      = 4
xlClipboardFormatDspText                  = 12
xlClipboardFormatEmbedSource              = 22
xlClipboardFormatEmbeddedObject           = 21
xlClipboardFormatLink                     = 11
xlClipboardFormatLinkSource               = 23
xlClipboardFormatLinkSourceDesc           = 32
xlClipboardFormatMovie                    = 24
xlClipboardFormatNative                   = 14
xlClipboardFormatObjectDesc               = 31
xlClipboardFormatObjectLink               = 19
xlClipboardFormatOwnerLink                = 17
xlClipboardFormatPICT                     = 2
xlClipboardFormatPrintPICT                = 3
xlClipboardFormatRTF                      = 7
xlClipboardFormatSYLK                     = 6
xlClipboardFormatScreenPICT               = 29
xlClipboardFormatStandardFont             = 28
xlClipboardFormatStandardScale            = 27
xlClipboardFormatTable                    = 16
xlClipboardFormatText                     = 0
xlClipboardFormatToolFace                 = 25
xlClipboardFormatToolFacePICT             = 26
xlClipboardFormatVALU                     = 1
xlClipboardFormatWK1                      = 10
xlCmdCube                                 = 1
xlCmdDAX                                  = 8
xlCmdDefault                              = 4
xlCmdExcel                                = 7
xlCmdList                                 = 5
xlCmdSql                                  = 2
xlCmdTable                                = 3
xlCmdTableCollection                      = 6
xlColorIndexAutomatic                     = -4105
xlColorIndexNone                          = -4142
xlDMYFormat                               = 4
xlDYMFormat                               = 7
xlEMDFormat                               = 10
xlGeneralFormat                           = 1
xlMDYFormat                               = 3
xlMYDFormat                               = 6
xlSkipColumn                              = 9
xlTextFormat                              = 2
xlYDMFormat                               = 8
xlYMDFormat                               = 5
xlCommandUnderlinesAutomatic              = -4105
xlCommandUnderlinesOff                    = -4146
xlCommandUnderlinesOn                     = 1
xlCommentAndIndicator                     = 1
xlCommentIndicatorOnly                    = -1
xlNoIndicator                             = 0
xlConditionValueAutomaticMax              = 7
xlConditionValueAutomaticMin              = 6
xlConditionValueFormula                   = 4
xlConditionValueHighestValue              = 2
xlConditionValueLowestValue               = 1
xlConditionValueNone                      = -1
xlConditionValueNumber                    = 0
xlConditionValuePercent                   = 3
xlConditionValuePercentile                = 5
xlConnectionTypeDATAFEED                  = 6
xlConnectionTypeMODEL                     = 7
xlConnectionTypeNOSOURCE                  = 9
xlConnectionTypeODBC                      = 2
xlConnectionTypeOLEDB                     = 1
xlConnectionTypeTEXT                      = 4
xlConnectionTypeWEB                       = 5
xlConnectionTypeWORKSHEET                 = 8
xlConnectionTypeXMLMAP                    = 3
xlAverage                                 = -4106
xlCount                                   = -4112
xlCountNums                               = -4113
xlDistinctCount                           = 11
xlMax                                     = -4136
xlMin                                     = -4139
xlProduct                                 = -4149
xlStDev                                   = -4155
xlStDevP                                  = -4156
xlSum                                     = -4157
xlUnknown                                 = 1000
xlVar                                     = -4164
xlVarP                                    = -4165
xlBeginsWith                              = 2
xlContains                                = 0
xlDoesNotContain                          = 1
xlEndsWith                                = 3
xlBitmap                                  = 2
xlPicture                                 = -4147
xlExtractData                             = 2
xlNormalLoad                              = 0
xlRepairFile                              = 1
xlCreatorCode                             = 1480803660
xlCredentialsMethodIntegrated             = 0
xlCredentialsMethodNone                   = 1
xlCredentialsMethodStored                 = 2
xlCubeAttribute                           = 4
xlCubeCalculatedMeasure                   = 5
xlCubeHierarchy                           = 1
xlCubeImplicitMeasure                     = 11
xlCubeKPIGoal                             = 7
xlCubeKPIStatus                           = 8
xlCubeKPITrend                            = 9
xlCubeKPIValue                            = 6
xlCubeKPIWeight                           = 10
xlCubeMeasure                             = 2
xlCubeSet                                 = 3
xlHierarchy                               = 1
xlMeasure                                 = 2
xlSet                                     = 3
xlCopy                                    = 1
xlCut                                     = 2
xlValidAlertInformation                   = 3
xlValidAlertStop                          = 1
xlValidAlertWarning                       = 2
xlValidateCustom                          = 7
xlValidateDate                            = 4
xlValidateDecimal                         = 2
xlValidateInputOnly                       = 0
xlValidateList                            = 3
xlValidateTextLength                      = 6
xlValidateTime                            = 5
xlValidateWholeNumber                     = 1
xlDataBarAxisAutomatic                    = 0
xlDataBarAxisMidpoint                     = 1
xlDataBarAxisNone                         = 2
xlDataBarBorderNone                       = 0
xlDataBarBorderSolid                      = 1
xlDataBarFillGradient                     = 1
xlDataBarFillSolid                        = 0
xlDataBarColor                            = 0
xlDataBarSameAsPositive                   = 1
xlLabelPositionAbove                      = 0
xlLabelPositionBelow                      = 1
xlLabelPositionBestFit                    = 5
xlLabelPositionCenter                     = -4108
xlLabelPositionCustom                     = 7
xlLabelPositionInsideBase                 = 4
xlLabelPositionInsideEnd                  = 3
xlLabelPositionLeft                       = -4131
xlLabelPositionMixed                      = 6
xlLabelPositionOutsideEnd                 = 2
xlLabelPositionRight                      = -4152
xlDataLabelSeparatorDefault               = 1
xlDataLabelsShowBubbleSizes               = 6
xlDataLabelsShowLabel                     = 4
xlDataLabelsShowLabelAndPercent           = 5
xlDataLabelsShowNone                      = -4142
xlDataLabelsShowPercent                   = 3
xlDataLabelsShowValue                     = 2
xlDay                                     = 1
xlMonth                                   = 3
xlWeekday                                 = 2
xlYear                                    = 4
xlAutoFill                                = 4
xlChronological                           = 3
xlDataSeriesLinear                        = -4132
xlGrowth                                  = 2
xlShiftToLeft                             = -4159
xlShiftUp                                 = -4162
xlDown                                    = -4121
xlToLeft                                  = -4159
xlToRight                                 = -4161
xlUp                                      = -4162
xlInterpolated                            = 3
xlNotPlotted                              = 1
xlZero                                    = 2
xlDisplayShapes                           = -4104
xlHide                                    = 3
xlPlaceholders                            = 2
xlHundredMillions                         = -8
xlHundredThousands                        = -5
xlHundreds                                = -2
xlMillionMillions                         = -10
xlMillions                                = -6
xlTenMillions                             = -7
xlTenThousands                            = -4
xlThousandMillions                        = -9
xlThousands                               = -3
xlDuplicate                               = 1
xlUnique                                  = 0
xlFilterAboveAverage                      = 33
xlFilterAllDatesInPeriodApril             = 24
xlFilterAllDatesInPeriodAugust            = 28
xlFilterAllDatesInPeriodDecember          = 32
xlFilterAllDatesInPeriodFebruray          = 22
xlFilterAllDatesInPeriodJanuary           = 21
xlFilterAllDatesInPeriodJuly              = 27
xlFilterAllDatesInPeriodJune              = 26
xlFilterAllDatesInPeriodMarch             = 23
xlFilterAllDatesInPeriodMay               = 25
xlFilterAllDatesInPeriodNovember          = 31
xlFilterAllDatesInPeriodOctober           = 30
xlFilterAllDatesInPeriodQuarter1          = 17
xlFilterAllDatesInPeriodQuarter2          = 18
xlFilterAllDatesInPeriodQuarter3          = 19
xlFilterAllDatesInPeriodQuarter4          = 20
xlFilterAllDatesInPeriodSeptember         = 29
xlFilterBelowAverage                      = 34
xlFilterLastMonth                         = 8
xlFilterLastQuarter                       = 11
xlFilterLastWeek                          = 5
xlFilterLastYear                          = 14
xlFilterNextMonth                         = 9
xlFilterNextQuarter                       = 12
xlFilterNextWeek                          = 6
xlFilterNextYear                          = 15
xlFilterThisMonth                         = 7
xlFilterThisQuarter                       = 10
xlFilterThisWeek                          = 4
xlFilterThisYear                          = 13
xlFilterToday                             = 1
xlFilterTomorrow                          = 3
xlFilterYearToDate                        = 16
xlFilterYesterday                         = 2
xlBIFF                                    = 2
xlPICT                                    = 1
xlRTF                                     = 4
xlVALU                                    = 8
xlAutomaticUpdate                         = 4
xlCancel                                  = 1
xlChangeAttributes                        = 6
xlManualUpdate                            = 5
xlOpenSource                              = 3
xlSelect                                  = 3
xlSendPublisher                           = 2
xlUpdateSubscriber                        = 2
xlPublisher                               = 1
xlSubscriber                              = 2
xlDisabled                                = 0
xlErrorHandler                            = 2
xlInterrupt                               = 1
xlNoRestrictions                          = 0
xlNoSelection                             = -4142
xlUnlockedCells                           = 1
xlCap                                     = 1
xlNoCap                                   = 2
xlX                                       = -4168
xlY                                       = 1
xlErrorBarIncludeBoth                     = 1
xlErrorBarIncludeMinusValues              = 3
xlErrorBarIncludeNone                     = -4142
xlErrorBarIncludePlusValues               = 2
xlErrorBarTypeCustom                      = -4114
xlErrorBarTypeFixedValue                  = 1
xlErrorBarTypePercent                     = 2
xlErrorBarTypeStDev                       = -4155
xlErrorBarTypeStError                     = 4
xlEmptyCellReferences                     = 7
xlEvaluateToError                         = 1
xlInconsistentFormula                     = 4
xlInconsistentListFormula                 = 9
xlListDataValidation                      = 8
xlMisleadingFormat                        = 10
xlNumberAsText                            = 3
xlOmittedCells                            = 5
xlTextDate                                = 2
xlUnlockedFormulaCells                    = 6
xlReadOnly                                = 3
xlReadWrite                               = 2
xlAddIn                                   = 18
xlAddIn8                                  = 18
xlCSV                                     = 6
xlCSVMSDOS                                = 24
xlCSVMac                                  = 22
xlCSVUTF8                                 = 62
xlCSVWindows                              = 23
xlCurrentPlatformText                     = -4158
xlDBF2                                    = 7
xlDBF3                                    = 8
xlDBF4                                    = 11
xlDIF                                     = 9
xlExcel12                                 = 50
xlExcel2                                  = 16
xlExcel2FarEast                           = 27
xlExcel3                                  = 29
xlExcel4                                  = 33
xlExcel4Workbook                          = 35
xlExcel5                                  = 39
xlExcel7                                  = 39
xlExcel8                                  = 56
xlExcel9795                               = 43
xlHtml                                    = 44
xlIntlAddIn                               = 26
xlIntlMacro                               = 25
xlOpenDocumentSpreadsheet                 = 60
xlOpenXMLAddIn                            = 55
xlOpenXMLStrictWorkbook                   = 61
xlOpenXMLTemplate                         = 54
xlOpenXMLTemplateMacroEnabled             = 53
xlOpenXMLWorkbook                         = 51
xlOpenXMLWorkbookMacroEnabled             = 52
xlSYLK                                    = 2
xlTemplate                                = 17
xlTemplate8                               = 17
xlTextMSDOS                               = 21
xlTextMac                                 = 19
xlTextPrinter                             = 36
xlTextWindows                             = 20
xlUnicodeText                             = 42
xlWJ2WD1                                  = 14
xlWJ3                                     = 40
xlWJ3FJ3                                  = 41
xlWK1                                     = 5
xlWK1ALL                                  = 31
xlWK1FMT                                  = 30
xlWK3                                     = 15
xlWK3FM3                                  = 32
xlWK4                                     = 38
xlWKS                                     = 4
xlWQ1                                     = 34
xlWebArchive                              = 45
xlWorkbookDefault                         = 51
xlWorkbookNormal                          = -4143
xlWorks2FarEast                           = 28
xlXMLSpreadsheet                          = 46
xlFileValidationPivotDefault              = 0
xlFileValidationPivotRun                  = 1
xlFileValidationPivotSkip                 = 2
xlFillWithAll                             = -4104
xlFillWithContents                        = 2
xlFillWithFormats                         = -4122
xlFilterCopy                              = 2
xlFilterInPlace                           = 1
xlFilterAllDatesInPeriodDay               = 2
xlFilterAllDatesInPeriodHour              = 3
xlFilterAllDatesInPeriodMinute            = 4
xlFilterAllDatesInPeriodMonth             = 1
xlFilterAllDatesInPeriodSecond            = 5
xlFilterAllDatesInPeriodYear              = 0
xlFilterStatusDateHasTime                 = 2
xlFilterStatusDateWrongOrder              = 1
xlFilterStatusInvalidDate                 = 3
xlFilterStatusOK                          = 0
xlComments                                = -4144
xlCommentsThreaded                        = -4184
xlFormulas                                = -4123
xlFormulas2                               = -4185
xlValues                                  = -4163
xlQualityMinimum                          = 1
xlQualityStandard                         = 0
xlTypePDF                                 = 0
xlTypeXPS                                 = 1
xlForecastAggregationAverage              = 1
xlForecastAggregationCount                = 2
xlForecastAggregationCountA               = 3
xlForecastAggregationMax                  = 4
xlForecastAggregationMedian               = 5
xlForecastAggregationMin                  = 6
xlForecastAggregationSum                  = 7
xlForecastChartTypeColumn                 = 1
xlForecastChartTypeLine                   = 0
xlForecastDataCompletionInterpolate       = 1
xlForecastDataCompletionZeros             = 0
xlButtonControl                           = 0
xlCheckBox                                = 1
xlDropDown                                = 2
xlEditBox                                 = 3
xlGroupBox                                = 4
xlLabel                                   = 5
xlListBox                                 = 6
xlOptionButton                            = 7
xlScrollBar                               = 8
xlSpinner                                 = 9
xlBetween                                 = 1
xlEqual                                   = 3
xlGreater                                 = 5
xlGreaterEqual                            = 7
xlLess                                    = 6
xlLessEqual                               = 8
xlNotBetween                              = 2
xlNotEqual                                = 4
xlAboveAverageCondition                   = 12
xlBlanksCondition                         = 10
xlCellValue                               = 1
xlColorScale                              = 3
xlDatabar                                 = 4
xlErrorsCondition                         = 16
xlExpression                              = 2
xlIconSets                                = 6
xlNoBlanksCondition                       = 13
xlNoErrorsCondition                       = 17
xlTextString                              = 9
xlTimePeriod                              = 11
xlTop10                                   = 5
xlUniqueValues                            = 8
xlFilterBottom                            = 0
xlFilterBottomPercent                     = 2
xlFilterTop                               = 1
xlFilterTopPercent                        = 3
xlColumnLabels                            = 2
xlMixedLabels                             = 3
xlNoLabels                                = -4142
xlRowLabels                               = 1
xlReplaceFormula                          = 0
xlReplaceFormula2                         = 1
xlGenerateTableRefA1                      = 0
xlGenerateTableRefStruct                  = 1
xlGeoMappingLevelAutomatic                = 0
xlGeoMappingLevelCountryRegion            = 5
xlGeoMappingLevelCountryRegionList        = 6
xlGeoMappingLevelCounty                   = 3
xlGeoMappingLevelDataOnly                 = 1
xlGeoMappingLevelPostalCode               = 2
xlGeoMappingLevelState                    = 4
xlGeoMappingLevelWorld                    = 7
xlGeoProjectionTypeAlbers                 = 3
xlGeoProjectionTypeAutomatic              = 0
xlGeoProjectionTypeMercator               = 1
xlGeoProjectionTypeMiller                 = 2
xlGeoProjectionTypeRobinson               = 4
xlGradientFillLinear                      = 0
xlGradientFillPath                        = 1
xlGradientStopPositionTypeExtremeValue    = 0
xlGradientStopPositionTypeNumber          = 1
xlGradientStopPositionTypePercent         = 2
xlHAlignCenter                            = -4108
xlHAlignCenterAcrossSelection             = 7
xlHAlignDistributed                       = -4117
xlHAlignFill                              = 5
xlHAlignGeneral                           = 1
xlHAlignJustify                           = -4130
xlHAlignLeft                              = -4131
xlHAlignRight                             = -4152
xlHebrewFullScript                        = 0
xlHebrewMixedAuthorizedScript             = 3
xlHebrewMixedScript                       = 2
xlHebrewPartialScript                     = 1
xlAllChanges                              = 2
xlNotYetReviewed                          = 3
xlSinceMyLastSave                         = 1
xlHtmlCalc                                = 1
xlHtmlChart                               = 3
xlHtmlList                                = 2
xlHtmlStatic                              = 0
xlIMEModeAlpha                            = 8
xlIMEModeAlphaFull                        = 7
xlIMEModeDisable                          = 3
xlIMEModeHangul                           = 10
xlIMEModeHangulFull                       = 9
xlIMEModeHiragana                         = 4
xlIMEModeKatakana                         = 5
xlIMEModeKatakanaHalf                     = 6
xlIMEModeNoControl                        = 0
xlIMEModeOff                              = 2
xlIMEModeOn                               = 1
xlIcon0Bars                               = 37
xlIcon0FilledBoxes                        = 52
xlIcon1Bar                                = 38
xlIcon1FilledBox                          = 51
xlIcon2Bars                               = 39
xlIcon2FilledBoxes                        = 50
xlIcon3Bars                               = 40
xlIcon3FilledBoxes                        = 49
xlIcon4Bars                               = 41
xlIcon4FilledBoxes                        = 48
xlIconBlackCircle                         = 32
xlIconBlackCircleWithBorder               = 13
xlIconCircleWithOneWhiteQuarter           = 33
xlIconCircleWithThreeWhiteQuarters        = 35
xlIconCircleWithTwoWhiteQuarters          = 34
xlIconGoldStar                            = 42
xlIconGrayCircle                          = 31
xlIconGrayDownArrow                       = 6
xlIconGrayDownInclineArrow                = 28
xlIconGraySideArrow                       = 5
xlIconGrayUpArrow                         = 4
xlIconGrayUpInclineArrow                  = 27
xlIconGreenCheck                          = 22
xlIconGreenCheckSymbol                    = 19
xlIconGreenCircle                         = 10
xlIconGreenFlag                           = 7
xlIconGreenTrafficLight                   = 14
xlIconGreenUpArrow                        = 1
xlIconGreenUpTriangle                     = 45
xlIconHalfGoldStar                        = 43
xlIconNoCellIcon                          = -1
xlIconPinkCircle                          = 30
xlIconRedCircle                           = 29
xlIconRedCircleWithBorder                 = 12
xlIconRedCross                            = 24
xlIconRedCrossSymbol                      = 21
xlIconRedDiamond                          = 18
xlIconRedDownArrow                        = 3
xlIconRedDownTriangle                     = 47
xlIconRedFlag                             = 9
xlIconRedTrafficLight                     = 16
xlIconSilverStar                          = 44
xlIconWhiteCircleAllWhiteQuarters         = 36
xlIconYellowCircle                        = 11
xlIconYellowDash                          = 46
xlIconYellowDownInclineArrow              = 26
xlIconYellowExclamation                   = 23
xlIconYellowExclamationSymbol             = 20
xlIconYellowFlag                          = 8
xlIconYellowSideArrow                     = 2
xlIconYellowTrafficLight                  = 15
xlIconYellowTriangle                      = 17
xlIconYellowUpInclineArrow                = 25
xl3Arrows                                 = 1
xl3ArrowsGray                             = 2
xl3Flags                                  = 3
xl3Signs                                  = 6
xl3Stars                                  = 18
xl3Symbols                                = 7
xl3Symbols2                               = 8
xl3TrafficLights1                         = 4
xl3TrafficLights2                         = 5
xl3Triangles                              = 19
xl4Arrows                                 = 9
xl4ArrowsGray                             = 10
xl4CRV                                    = 12
xl4RedToBlack                             = 11
xl4TrafficLights                          = 13
xl5Arrows                                 = 14
xl5ArrowsGray                             = 15
xl5Boxes                                  = 20
xl5CRV                                    = 16
xl5Quarters                               = 17
xlCustomSet                               = -1
xlPivotTableReport                        = 1
xlQueryTable                              = 0
xlTable                                   = 2
xlFormatFromLeftOrAbove                   = 0
xlFormatFromRightOrBelow                  = 1
xlShiftDown                               = -4121
xlShiftToRight                            = -4161
xlOutline                                 = 1
xlTabular                                 = 0
xlCompactRow                              = 0
xlOutlineRow                              = 2
xlTabularRow                              = 1
xlLegendPositionBottom                    = -4107
xlLegendPositionCorner                    = 2
xlLegendPositionCustom                    = -4161
xlLegendPositionLeft                      = -4131
xlLegendPositionRight                     = -4152
xlLegendPositionTop                       = -4160
xlContinuous                              = 1
xlDash                                    = -4115
xlDashDot                                 = 4
xlDashDotDot                              = 5
xlDot                                     = -4118
xlDouble                                  = -4119
xlLineStyleNone                           = -4142
xlSlantDashDot                            = 13
xlExcelLinks                              = 1
xlOLELinks                                = 2
xlPublishers                              = 5
xlSubscribers                             = 6
xlEditionDate                             = 2
xlLinkInfoStatus                          = 3
xlUpdateState                             = 1
xlLinkInfoOLELinks                        = 2
xlLinkInfoPublishers                      = 5
xlLinkInfoSubscribers                     = 6
xlLinkStatusCopiedValues                  = 10
xlLinkStatusIndeterminate                 = 5
xlLinkStatusInvalidName                   = 7
xlLinkStatusMissingFile                   = 1
xlLinkStatusMissingSheet                  = 2
xlLinkStatusNotStarted                    = 6
xlLinkStatusOK                            = 0
xlLinkStatusOld                           = 3
xlLinkStatusSourceNotCalculated           = 4
xlLinkStatusSourceNotOpen                 = 8
xlLinkStatusSourceOpen                    = 9
xlLinkTypeExcelLinks                      = 1
xlLinkTypeOLELinks                        = 2
xlLinkedDataTypeStateBrokenLinkedData     = 3
xlLinkedDataTypeStateDisambiguationNeeded = 2
xlLinkedDataTypeStateFetchingData         = 4
xlLinkedDataTypeStateNone                 = 0
xlLinkedDataTypeStateValidLinkedData      = 1
xlListConflictDialog                      = 0
xlListConflictDiscardAllConflicts         = 2
xlListConflictError                       = 3
xlListConflictRetryAllConflicts           = 1
xlListDataTypeCheckbox                    = 9
xlListDataTypeChoice                      = 6
xlListDataTypeChoiceMulti                 = 7
xlListDataTypeCounter                     = 11
xlListDataTypeCurrency                    = 4
xlListDataTypeDateTime                    = 5
xlListDataTypeHyperLink                   = 10
xlListDataTypeListLookup                  = 8
xlListDataTypeMultiLineRichText           = 12
xlListDataTypeMultiLineText               = 2
xlListDataTypeNone                        = 0
xlListDataTypeNumber                      = 3
xlListDataTypeText                        = 1
xlSrcExternal                             = 0
xlSrcModel                                = 4
xlSrcQuery                                = 3
xlSrcRange                                = 1
xlSrcXml                                  = 2
xlColumnHeader                            = -4110
xlColumnItem                              = 5
xlDataHeader                              = 3
xlDataItem                                = 7
xlPageHeader                              = 2
xlPageItem                                = 6
xlRowHeader                               = -4153
xlRowItem                                 = 4
xlTableBody                               = 8
xlPart                                    = 2
xlWhole                                   = 1
xlLookForBlanks                           = 0
xlLookForErrors                           = 1
xlLookForFormulas                         = 2
xlMicrosoftAccess                         = 4
xlMicrosoftFoxPro                         = 5
xlMicrosoftMail                           = 3
xlMicrosoftPowerPoint                     = 2
xlMicrosoftProject                        = 6
xlMicrosoftSchedulePlus                   = 7
xlMicrosoftWord                           = 1
xlMAPI                                    = 1
xlNoMailSystem                            = 0
xlPowerTalk                               = 2
xlMarkerStyleAutomatic                    = -4105
xlMarkerStyleCircle                       = 8
xlMarkerStyleDash                         = -4115
xlMarkerStyleDiamond                      = 2
xlMarkerStyleDot                          = -4118
xlMarkerStyleNone                         = -4142
xlMarkerStylePicture                      = -4147
xlMarkerStylePlus                         = 9
xlMarkerStyleSquare                       = 1
xlMarkerStyleStar                         = 5
xlMarkerStyleTriangle                     = 3
xlMarkerStyleX                            = -4168
xlCentimeters                             = 1
xlInches                                  = 0
xlMillimeters                             = 2
xlChangeByExcel                           = 0
xlChangeByPowerPivotAddIn                 = 1
xlNoButton                                = 0
xlPrimaryButton                           = 1
xlSecondaryButton                         = 2
xlDefault                                 = -4143
xlIBeam                                   = 3
xlNorthwestArrow                          = 1
xlWait                                    = 2
xlOLEControl                              = 2
xlOLEEmbed                                = 1
xlOLELink                                 = 0
xlVerbOpen                                = 2
xlVerbPrimary                             = 1
xlOartHorizontalOverflowClip              = 1
xlOartHorizontalOverflowOverflow          = 0
xlOartVerticalOverflowClip                = 1
xlOartVerticalOverflowEllipsis            = 2
xlOartVerticalOverflowOverflow            = 0
xlFitToPage                               = 2
xlFullPage                                = 3
xlScreenSize                              = 1
xlDownThenOver                            = 1
xlOverThenDown                            = 2
xlDownward                                = -4170
xlHorizontal                              = -4128
xlUpward                                  = -4171
xlVertical                                = -4166
xlBlanks                                  = 4
xlButton                                  = 15
xlDataAndLabel                            = 0
xlDataOnly                                = 2
xlFirstRow                                = 256
xlLabelOnly                               = 1
xlOrigin                                  = 3
xlPageBreakAutomatic                      = -4105
xlPageBreakManual                         = -4135
xlPageBreakNone                           = -4142
xlPageBreakFull                           = 1
xlPageBreakPartial                        = 2
xlLandscape                               = 2
xlPortrait                                = 1
xlPaper10x14                              = 16
xlPaper11x17                              = 17
xlPaperA3                                 = 8
xlPaperA4                                 = 9
xlPaperA4Small                            = 10
xlPaperA5                                 = 11
xlPaperB4                                 = 12
xlPaperB5                                 = 13
xlPaperCsheet                             = 24
xlPaperDsheet                             = 25
xlPaperEnvelope10                         = 20
xlPaperEnvelope11                         = 21
xlPaperEnvelope12                         = 22
xlPaperEnvelope14                         = 23
xlPaperEnvelope9                          = 19
xlPaperEnvelopeB4                         = 33
xlPaperEnvelopeB5                         = 34
xlPaperEnvelopeB6                         = 35
xlPaperEnvelopeC3                         = 29
xlPaperEnvelopeC4                         = 30
xlPaperEnvelopeC5                         = 28
xlPaperEnvelopeC6                         = 31
xlPaperEnvelopeC65                        = 32
xlPaperEnvelopeDL                         = 27
xlPaperEnvelopeItaly                      = 36
xlPaperEnvelopeMonarch                    = 37
xlPaperEnvelopePersonal                   = 38
xlPaperEsheet                             = 26
xlPaperExecutive                          = 7
xlPaperFanfoldLegalGerman                 = 41
xlPaperFanfoldStdGerman                   = 40
xlPaperFanfoldUS                          = 39
xlPaperFolio                              = 14
xlPaperLedger                             = 4
xlPaperLegal                              = 5
xlPaperLetter                             = 1
xlPaperLetterSmall                        = 2
xlPaperNote                               = 18
xlPaperQuarto                             = 15
xlPaperStatement                          = 6
xlPaperTabloid                            = 3
xlPaperUser                               = 256
xlParamTypeBigInt                         = -5
xlParamTypeBinary                         = -2
xlParamTypeBit                            = -7
xlParamTypeChar                           = 1
xlParamTypeDate                           = 9
xlParamTypeDecimal                        = 3
xlParamTypeDouble                         = 8
xlParamTypeFloat                          = 6
xlParamTypeInteger                        = 4
xlParamTypeLongVarBinary                  = -4
xlParamTypeLongVarChar                    = -1
xlParamTypeNumeric                        = 2
xlParamTypeReal                           = 7
xlParamTypeSmallInt                       = 5
xlParamTypeTime                           = 10
xlParamTypeTimestamp                      = 11
xlParamTypeTinyInt                        = -6
xlParamTypeUnknown                        = 0
xlParamTypeVarBinary                      = -3
xlParamTypeVarChar                        = 12
xlParamTypeWChar                          = -8
xlConstant                                = 1
xlPrompt                                  = 0
xlRange                                   = 2
xlParentDataLabelOptionsBanner            = 1
xlParentDataLabelOptionsNone              = 0
xlParentDataLabelOptionsOverlapping       = 2
xlPasteSpecialOperationAdd                = 2
xlPasteSpecialOperationDivide             = 5
xlPasteSpecialOperationMultiply           = 4
xlPasteSpecialOperationNone               = -4142
xlPasteSpecialOperationSubtract           = 3
xlPasteAll                                = -4104
xlPasteAllExceptBorders                   = 7
xlPasteAllMergingConditionalFormats       = 14
xlPasteAllUsingSourceTheme                = 13
xlPasteColumnWidths                       = 8
xlPasteComments                           = -4144
xlPasteFormats                            = -4122
xlPasteFormulas                           = -4123
xlPasteFormulasAndNumberFormats           = 11
xlPasteValidation                         = 6
xlPasteValues                             = -4163
xlPasteValuesAndNumberFormats             = 12
xlPatternAutomatic                        = -4105
xlPatternChecker                          = 9
xlPatternCrissCross                       = 16
xlPatternDown                             = -4121
xlPatternGray16                           = 17
xlPatternGray25                           = -4124
xlPatternGray50                           = -4125
xlPatternGray75                           = -4126
xlPatternGray8                            = 18
xlPatternGrid                             = 15
xlPatternHorizontal                       = -4128
xlPatternLightDown                        = 13
xlPatternLightHorizontal                  = 11
xlPatternLightUp                          = 14
xlPatternLightVertical                    = 12
xlPatternLinearGradient                   = 4000
xlPatternNone                             = -4142
xlPatternRectangularGradient              = 4001
xlPatternSemiGray75                       = 10
xlPatternSolid                            = 1
xlPatternUp                               = -4162
xlPatternVertical                         = -4166
xlPhoneticAlignCenter                     = 2
xlPhoneticAlignDistributed                = 3
xlPhoneticAlignLeft                       = 1
xlPhoneticAlignNoControl                  = 0
xlHiragana                                = 2
xlKatakana                                = 1
xlKatakanaHalf                            = 0
xlNoConversion                            = 3
xlPrinter                                 = 2
xlScreen                                  = 1
xlBMP                                     = 1
xlCGM                                     = 7
xlDRW                                     = 4
xlDXF                                     = 5
xlEPS                                     = 8
xlHGL                                     = 6
xlPCT                                     = 13
xlPCX                                     = 10
xlPIC                                     = 11
xlPLT                                     = 12
xlTIF                                     = 9
xlWMF                                     = 2
xlWPG                                     = 3
xlCenterPoint                             = 5
xlInnerCenterPoint                        = 8
xlInnerClockwisePoint                     = 7
xlInnerCounterClockwisePoint              = 9
xlMidClockwiseRadiusPoint                 = 4
xlMidCounterClockwiseRadiusPoint          = 6
xlOuterCenterPoint                        = 2
xlOuterClockwisePoint                     = 3
xlOuterCounterClockwisePoint              = 1
xlHorizontalCoordinate                    = 1
xlVerticalCoordinate                      = 2
xlPivotCellBlankCell                      = 9
xlPivotCellCustomSubtotal                 = 7
xlPivotCellDataField                      = 4
xlPivotCellDataPivotField                 = 8
xlPivotCellGrandTotal                     = 3
xlPivotCellPageFieldItem                  = 6
xlPivotCellPivotField                     = 5
xlPivotCellPivotItem                      = 1
xlPivotCellSubtotal                       = 2
xlPivotCellValue                          = 0
xlDataFieldScope                          = 2
xlFieldsScope                             = 1
xlSelectionScope                          = 0
xlDifferenceFrom                          = 2
xlIndex                                   = 9
xlNoAdditionalCalculation                 = -4143
xlPercentDifferenceFrom                   = 4
xlPercentOf                               = 3
xlPercentOfColumn                         = 7
xlPercentOfParent                         = 12
xlPercentOfParentColumn                   = 11
xlPercentOfParentRow                      = 10
xlPercentOfRow                            = 6
xlPercentOfTotal                          = 8
xlPercentRunningTotal                     = 13
xlRankAscending                           = 14
xlRankDecending                           = 15
xlRunningTotal                            = 5
xlDate                                    = 2
xlNumber                                  = -4145
xlText                                    = -4158
xlColumnField                             = 2
xlDataField                               = 4
xlHidden                                  = 0
xlPageField                               = 3
xlRowField                                = 1
xlDoNotRepeatLabels                       = 1
xlRepeatLabels                            = 2
xlAfter                                   = 33
xlAfterOrEqualTo                          = 34
xlAllDatesInPeriodApril                   = 60
xlAllDatesInPeriodAugust                  = 64
xlAllDatesInPeriodDecember                = 68
xlAllDatesInPeriodFebruary                = 58
xlAllDatesInPeriodJanuary                 = 57
xlAllDatesInPeriodJuly                    = 63
xlAllDatesInPeriodJune                    = 62
xlAllDatesInPeriodMarch                   = 59
xlAllDatesInPeriodMay                     = 61
xlAllDatesInPeriodNovember                = 67
xlAllDatesInPeriodOctober                 = 66
xlAllDatesInPeriodQuarter1                = 53
xlAllDatesInPeriodQuarter2                = 54
xlAllDatesInPeriodQuarter3                = 55
xlAllDatesInPeriodQuarter4                = 56
xlAllDatesInPeriodSeptember               = 65
xlBefore                                  = 31
xlBeforeOrEqualTo                         = 32
xlBottomCount                             = 2
xlBottomPercent                           = 4
xlBottomSum                               = 6
xlCaptionBeginsWith                       = 17
xlCaptionContains                         = 21
xlCaptionDoesNotBeginWith                 = 18
xlCaptionDoesNotContain                   = 22
xlCaptionDoesNotEndWith                   = 20
xlCaptionDoesNotEqual                     = 16
xlCaptionEndsWith                         = 19
xlCaptionEquals                           = 15
xlCaptionIsBetween                        = 27
xlCaptionIsGreaterThan                    = 23
xlCaptionIsGreaterThanOrEqualTo           = 24
xlCaptionIsLessThan                       = 25
xlCaptionIsLessThanOrEqualTo              = 26
xlCaptionIsNotBetween                     = 28
xlDateBetween                             = 35
xlDateLastMonth                           = 45
xlDateLastQuarter                         = 48
xlDateLastWeek                            = 42
xlDateLastYear                            = 51
xlDateNextMonth                           = 43
xlDateNextQuarter                         = 46
xlDateNextWeek                            = 40
xlDateNextYear                            = 49
xlDateNotBetween                          = 36
xlDateThisMonth                           = 44
xlDateThisQuarter                         = 47
xlDateThisWeek                            = 41
xlDateThisYear                            = 50
xlDateToday                               = 38
xlDateTomorrow                            = 37
xlDateYesterday                           = 39
xlNotSpecificDate                         = 30
xlSpecificDate                            = 29
xlTopCount                                = 1
xlTopPercent                              = 3
xlTopSum                                  = 5
xlValueDoesNotEqual                       = 8
xlValueEquals                             = 7
xlValueIsBetween                          = 13
xlValueIsGreaterThan                      = 9
xlValueIsGreaterThanOrEqualTo             = 10
xlValueIsLessThan                         = 11
xlValueIsLessThanOrEqualTo                = 12
xlValueIsNotBetween                       = 14
xlYearToDate                              = 52
xlPTClassic                               = 20
xlPTNone                                  = 21
xlReport1                                 = 0
xlReport10                                = 9
xlReport2                                 = 1
xlReport3                                 = 2
xlReport4                                 = 3
xlReport5                                 = 4
xlReport6                                 = 5
xlReport7                                 = 6
xlReport8                                 = 7
xlReport9                                 = 8
xlTable1                                  = 10
xlTable10                                 = 19
xlTable2                                  = 11
xlTable3                                  = 12
xlTable4                                  = 13
xlTable5                                  = 14
xlTable6                                  = 15
xlTable7                                  = 16
xlTable8                                  = 17
xlTable9                                  = 18
xlPivotLineBlank                          = 3
xlPivotLineGrandTotal                     = 2
xlPivotLineRegular                        = 0
xlPivotLineSubtotal                       = 1
xlMissingItemsDefault                     = -1
xlMissingItemsMax                         = 32500
xlMissingItemsMax2                        = 1048576
xlMissingItemsNone                        = 0
xlConsolidation                           = 3
xlDatabase                                = 1
xlExternal                                = 2
xlPivotTable                              = -4148
xlScenario                                = 4
xlPivotTableVersion10                     = 1
xlPivotTableVersion11                     = 2
xlPivotTableVersion12                     = 3
xlPivotTableVersion14                     = 4
xlPivotTableVersion15                     = 5
xlPivotTableVersion2000                   = 0
xlPivotTableVersionCurrent                = -1
xlFreeFloating                            = 3
xlMove                                    = 2
xlMoveAndSize                             = 1
xlMSDOS                                   = 3
xlMacintosh                               = 1
xlWindows                                 = 2
xlPortugueseBoth                          = 3
xlPortuguesePostReform                    = 2
xlPortuguesePreReform                     = 1
xlPrintErrorsBlank                        = 1
xlPrintErrorsDash                         = 2
xlPrintErrorsDisplayed                    = 0
xlPrintErrorsNA                           = 3
xlPrintInPlace                            = 16
xlPrintNoComments                         = -4142
xlPrintSheetEnd                           = 1
xlPriorityHigh                            = -4127
xlPriorityLow                             = -4134
xlPriorityNormal                          = -4143
xlDisplayPropertyInPivotTable             = 1
xlDisplayPropertyInPivotTableAndTooltip   = 3
xlDisplayPropertyInTooltip                = 2
xlProtectedViewCloseEdit                  = 1
xlProtectedViewCloseForced                = 2
xlProtectedViewCloseNormal                = 0
xlProtectedViewWindowMaximized            = 2
xlProtectedViewWindowMinimized            = 1
xlProtectedViewWindowNormal               = 0
msoLimited                                = 1
msoNoOverwrite                            = 3
msoOrganization                           = 2
msoPublic                                 = 0
msoPBIAbort                               = 1
msoPBIIgnore                              = 0
msoPBIOverwrite                           = 2
msoPBIExport                              = 0
msoPBIUpload                              = 1
xlADORecordset                            = 7
xlDAORecordset                            = 2
xlODBCQuery                               = 1
xlOLEDBQuery                              = 5
xlTextImport                              = 6
xlWebQuery                                = 4
xlFormatConditions                        = 1
xlLensOnly                                = 0
xlRecommendedCharts                       = 2
xlSparklines                              = 5
xlTables                                  = 4
xlTotals                                  = 3
xlRangeAutoFormat3DEffects1               = 13
xlRangeAutoFormat3DEffects2               = 14
xlRangeAutoFormatAccounting1              = 4
xlRangeAutoFormatAccounting2              = 5
xlRangeAutoFormatAccounting3              = 6
xlRangeAutoFormatAccounting4              = 17
xlRangeAutoFormatClassic1                 = 1
xlRangeAutoFormatClassic2                 = 2
xlRangeAutoFormatClassic3                 = 3
xlRangeAutoFormatClassicPivotTable        = 31
xlRangeAutoFormatColor1                   = 7
xlRangeAutoFormatColor2                   = 8
xlRangeAutoFormatColor3                   = 9
xlRangeAutoFormatList1                    = 10
xlRangeAutoFormatList2                    = 11
xlRangeAutoFormatList3                    = 12
xlRangeAutoFormatLocalFormat1             = 15
xlRangeAutoFormatLocalFormat2             = 16
xlRangeAutoFormatLocalFormat3             = 19
xlRangeAutoFormatLocalFormat4             = 20
xlRangeAutoFormatNone                     = -4142
xlRangeAutoFormatPTNone                   = 42
xlRangeAutoFormatReport1                  = 21
xlRangeAutoFormatReport10                 = 30
xlRangeAutoFormatReport2                  = 22
xlRangeAutoFormatReport3                  = 23
xlRangeAutoFormatReport4                  = 24
xlRangeAutoFormatReport5                  = 25
xlRangeAutoFormatReport6                  = 26
xlRangeAutoFormatReport7                  = 27
xlRangeAutoFormatReport8                  = 28
xlRangeAutoFormatReport9                  = 29
xlRangeAutoFormatSimple                   = -4154
xlRangeAutoFormatTable1                   = 32
xlRangeAutoFormatTable10                  = 41
xlRangeAutoFormatTable2                   = 33
xlRangeAutoFormatTable3                   = 34
xlRangeAutoFormatTable4                   = 35
xlRangeAutoFormatTable5                   = 36
xlRangeAutoFormatTable6                   = 37
xlRangeAutoFormatTable7                   = 38
xlRangeAutoFormatTable8                   = 39
xlRangeAutoFormatTable9                   = 40
xlRangeValueDefault                       = 10
xlRangeValueMSPersistXML                  = 12
xlRangeValueXMLSpreadsheet                = 11
xlA1                                      = 1
xlR1C1                                    = -4150
xlAbsRowRelColumn                         = 2
xlAbsolute                                = 1
xlRelRowAbsColumn                         = 3
xlRelative                                = 4
xlRegionLabelOptionsBestFitOnly           = 1
xlRegionLabelOptionsNone                  = 0
xlRegionLabelOptionsShowAll               = 2
xlRDIAll                                  = 99
xlRDIComments                             = 1
xlRDIContentType                          = 16
xlRDIDefinedNameComments                  = 18
xlRDIDocumentManagementPolicy             = 15
xlRDIDocumentProperties                   = 8
xlRDIDocumentServerProperties             = 14
xlRDIDocumentWorkspace                    = 10
xlRDIEmailHeader                          = 5
xlRDIExcelDataModel                       = 23
xlRDIInactiveDataConnections              = 19
xlRDIInkAnnotations                       = 11
xlRDIInlineWebExtensions                  = 21
xlRDIPrinterPath                          = 20
xlRDIPublishInfo                          = 13
xlRDIRemovePersonalInformation            = 4
xlRDIRoutingSlip                          = 6
xlRDIScenarioComments                     = 12
xlRDISendForReview                        = 7
xlRDITaskpaneWebExtensions                = 22
rgbAliceBlue                              = 16775408
rgbAntiqueWhite                           = 14150650
rgbAqua                                   = 16776960
rgbAquamarine                             = 13959039
rgbAzure                                  = 16777200
rgbBeige                                  = 14480885
rgbBisque                                 = 12903679
rgbBlack                                  = 0
rgbBlanchedAlmond                         = 13495295
rgbBlue                                   = 16711680
rgbBlueViolet                             = 14822282
rgbBrown                                  = 2763429
rgbBurlyWood                              = 8894686
rgbCadetBlue                              = 10526303
rgbChartreuse                             = 65407
rgbCoral                                  = 5275647
rgbCornflowerBlue                         = 15570276
rgbCornsilk                               = 14481663
rgbCrimson                                = 3937500
rgbDarkBlue                               = 9109504
rgbDarkCyan                               = 9145088
rgbDarkGoldenrod                          = 755384
rgbDarkGray                               = 11119017
rgbDarkGreen                              = 25600
rgbDarkGrey                               = 11119017
rgbDarkKhaki                              = 7059389
rgbDarkMagenta                            = 9109643
rgbDarkOliveGreen                         = 3107669
rgbDarkOrange                             = 36095
rgbDarkOrchid                             = 13382297
rgbDarkRed                                = 139
rgbDarkSalmon                             = 8034025
rgbDarkSeaGreen                           = 9419919
rgbDarkSlateBlue                          = 9125192
rgbDarkSlateGray                          = 5197615
rgbDarkSlateGrey                          = 5197615
rgbDarkTurquoise                          = 13749760
rgbDarkViolet                             = 13828244
rgbDeepPink                               = 9639167
rgbDeepSkyBlue                            = 16760576
rgbDimGray                                = 6908265
rgbDimGrey                                = 6908265
rgbDodgerBlue                             = 16748574
rgbFireBrick                              = 2237106
rgbFloralWhite                            = 15792895
rgbForestGreen                            = 2263842
rgbFuchsia                                = 16711935
rgbGainsboro                              = 14474460
rgbGhostWhite                             = 16775416
rgbGold                                   = 55295
rgbGoldenrod                              = 2139610
rgbGray                                   = 8421504
rgbGreen                                  = 32768
rgbGreenYellow                            = 3145645
rgbGrey                                   = 8421504
rgbHoneydew                               = 15794160
rgbHotPink                                = 11823615
rgbIndianRed                              = 6053069
rgbIndigo                                 = 8519755
rgbIvory                                  = 15794175
rgbKhaki                                  = 9234160
rgbLavender                               = 16443110
rgbLavenderBlush                          = 16118015
rgbLawnGreen                              = 64636
rgbLemonChiffon                           = 13499135
rgbLightBlue                              = 15128749
rgbLightCoral                             = 8421616
rgbLightCyan                              = 9145088
rgbLightGoldenrodYellow                   = 13826810
rgbLightGray                              = 13882323
rgbLightGreen                             = 9498256
rgbLightGrey                              = 13882323
rgbLightPink                              = 12695295
rgbLightSalmon                            = 8036607
rgbLightSeaGreen                          = 11186720
rgbLightSkyBlue                           = 16436871
rgbLightSlateGray                         = 10061943
rgbLightSlateGrey                         = 10061943
rgbLightSteelBlue                         = 14599344
rgbLightYellow                            = 14745599
rgbLime                                   = 65280
rgbLimeGreen                              = 3329330
rgbLinen                                  = 15134970
rgbMaroon                                 = 128
rgbMediumAquamarine                       = 11206502
rgbMediumBlue                             = 13434880
rgbMediumOrchid                           = 13850042
rgbMediumPurple                           = 14381203
rgbMediumSeaGreen                         = 7451452
rgbMediumSlateBlue                        = 15624315
rgbMediumSpringGreen                      = 10156544
rgbMediumTurquoise                        = 13422920
rgbMediumVioletRed                        = 8721863
rgbMidnightBlue                           = 7346457
rgbMintCream                              = 16449525
rgbMistyRose                              = 14804223
rgbMoccasin                               = 11920639
rgbNavajoWhite                            = 11394815
rgbNavy                                   = 8388608
rgbNavyBlue                               = 8388608
rgbOldLace                                = 15136253
rgbOlive                                  = 32896
rgbOliveDrab                              = 2330219
rgbOrange                                 = 42495
rgbOrangeRed                              = 17919
rgbOrchid                                 = 14053594
rgbPaleGoldenrod                          = 7071982
rgbPaleGreen                              = 10025880
rgbPaleTurquoise                          = 15658671
rgbPaleVioletRed                          = 9662683
rgbPapayaWhip                             = 14020607
rgbPeachPuff                              = 12180223
rgbPeru                                   = 4163021
rgbPink                                   = 13353215
rgbPlum                                   = 14524637
rgbPowderBlue                             = 15130800
rgbPurple                                 = 8388736
rgbRed                                    = 255
rgbRosyBrown                              = 9408444
rgbRoyalBlue                              = 14772545
rgbSalmon                                 = 7504122
rgbSandyBrown                             = 6333684
rgbSeaGreen                               = 5737262
rgbSeashell                               = 15660543
rgbSienna                                 = 2970272
rgbSilver                                 = 12632256
rgbSkyBlue                                = 15453831
rgbSlateBlue                              = 13458026
rgbSlateGray                              = 9470064
rgbSlateGrey                              = 9470064
rgbSnow                                   = 16448255
rgbSpringGreen                            = 8388352
rgbSteelBlue                              = 11829830
rgbTan                                    = 9221330
rgbTeal                                   = 8421376
rgbThistle                                = 14204888
rgbTomato                                 = 4678655
rgbTurquoise                              = 13688896
rgbViolet                                 = 15631086
rgbWheat                                  = 11788021
rgbWhite                                  = 16777215
rgbWhiteSmoke                             = 16119285
rgbYellow                                 = 65535
rgbYellowGreen                            = 3329434
xlAlways                                  = 1
xlAsRequired                              = 0
xlNever                                   = 2
xlAllAtOnce                               = 2
xlOneAfterAnother                         = 1
xlNotYetRouted                            = 0
xlRoutingComplete                         = 2
xlRoutingInProgress                       = 1
xlColumns                                 = 2
xlRows                                    = 1
xlAutoActivate                            = 3
xlAutoClose                               = 2
xlAutoDeactivate                          = 4
xlAutoOpen                                = 1
xlDoNotSaveChanges                        = 2
xlSaveChanges                             = 1
xlExclusive                               = 3
xlNoChange                                = 1
xlShared                                  = 2
xlLocalSessionChanges                     = 2
xlOtherSessionChanges                     = 3
xlUserResolution                          = 1
xlScaleLinear                             = -4132
xlScaleLogarithmic                        = -4133
xlNext                                    = 1
xlPrevious                                = 2
xlByColumns                               = 2
xlByRows                                  = 1
xlWithinSheet                             = 1
xlWithinWorkbook                          = 2
xlSeriesColorGradientStyleDiverging       = 1
xlSeriesColorGradientStyleSequential      = 0
xlSeriesNameLevelAll                      = -1
xlSeriesNameLevelCustom                   = -2
xlSeriesNameLevelNone                     = -3
xlChart                                   = -4109
xlDialogSheet                             = -4116
xlExcel4IntlMacroSheet                    = 4
xlExcel4MacroSheet                        = 3
xlWorksheet                               = -4167
xlSheetHidden                             = 0
xlSheetVeryHidden                         = 2
xlSheetVisible                            = -1
xlSizeIsArea                              = 1
xlSizeIsWidth                             = 2
xlSlicer                                  = 1
xlTimeline                                = 2
xlSlicerCrossFilterHideButtonsWithNoData  = 4
xlSlicerCrossFilterShowItemsWithDataAtTop = 2
xlSlicerCrossFilterShowItemsWithNoData    = 3
xlSlicerNoCrossFilter                     = 1
xlSlicerSortAscending                     = 2
xlSlicerSortDataSourceOrder               = 1
xlSlicerSortDescending                    = 3
xlSmartTagControlActiveX                  = 13
xlSmartTagControlButton                   = 6
xlSmartTagControlCheckbox                 = 9
xlSmartTagControlCombo                    = 12
xlSmartTagControlHelp                     = 3
xlSmartTagControlHelpURL                  = 4
xlSmartTagControlImage                    = 8
xlSmartTagControlLabel                    = 7
xlSmartTagControlLink                     = 2
xlSmartTagControlListbox                  = 11
xlSmartTagControlRadioGroup               = 14
xlSmartTagControlSeparator                = 5
xlSmartTagControlSmartTag                 = 1
xlSmartTagControlTextbox                  = 10
xlButtonOnly                              = 2
xlDisplayNone                             = 1
xlIndicatorAndButton                      = 0
xlSortNormal                              = 0
xlSortTextAsNumbers                       = 1
xlPinYin                                  = 1
xlStroke                                  = 2
xlCodePage                                = 2
xlSyllabary                               = 1
xlSortOnCellColor                         = 1
xlSortOnFontColor                         = 2
xlSortOnIcon                              = 3
xlSortOnValues                            = 0
xlAscending                               = 1
xlDescending                              = 2
xlSortColumns                             = 1
xlSortRows                                = 2
xlSortLabels                              = 2
xlSortValues                              = 1
xlSourceAutoFilter                        = 3
xlSourceChart                             = 5
xlSourcePivotTable                        = 6
xlSourcePrintArea                         = 2
xlSourceQuery                             = 7
xlSourceRange                             = 4
xlSourceSheet                             = 1
xlSourceWorkbook                          = 0
xlSpanishTuteoAndVoseo                    = 1
xlSpanishTuteoOnly                        = 0
xlSpanishVoseoOnly                        = 2
xlSparkScaleCustom                        = 3
xlSparkScaleGroup                         = 1
xlSparkScaleSingle                        = 2
xlSparkColumn                             = 2
xlSparkColumnStacked100                   = 3
xlSparkLine                               = 1
xlSparklineColumnsSquare                  = 2
xlSparklineNonSquare                      = 0
xlSparklineRowsSquare                     = 1
xlSpeakByColumns                          = 1
xlSpeakByRows                             = 0
xlErrors                                  = 16
xlLogical                                 = 4
xlNumbers                                 = 1
xlTextValues                              = 2
xlColorScaleBlackWhite                    = 3
xlColorScaleGYR                           = 2
xlColorScaleRYG                           = 1
xlColorScaleWhiteBlack                    = 4
xlSubscribeToPicture                      = -4147
xlSubscribeToText                         = -4158
xlAtBottom                                = 2
xlAtTop                                   = 1
xlSummaryOnLeft                           = -4131
xlSummaryOnRight                          = -4152
xlStandardSummary                         = 1
xlSummaryPivotTable                       = -4148
xlSummaryAbove                            = 0
xlSummaryBelow                            = 1
xlTabPositionFirst                        = 0
xlTabPositionLast                         = 1
xlBlankRow                                = 19
xlColumnStripe1                           = 7
xlColumnStripe2                           = 8
xlColumnSubheading1                       = 20
xlColumnSubheading2                       = 21
xlColumnSubheading3                       = 22
xlFirstColumn                             = 3
xlFirstHeaderCell                         = 9
xlFirstTotalCell                          = 11
xlGrandTotalColumn                        = 4
xlGrandTotalRow                           = 2
xlHeaderRow                               = 1
xlLastColumn                              = 4
xlLastHeaderCell                          = 10
xlLastTotalCell                           = 12
xlPageFieldLabels                         = 26
xlPageFieldValues                         = 27
xlRowStripe1                              = 5
xlRowStripe2                              = 6
xlRowSubheading1                          = 23
xlRowSubheading2                          = 24
xlRowSubheading3                          = 25
xlSlicerHoveredSelectedItemWithData       = 33
xlSlicerHoveredSelectedItemWithNoData     = 35
xlSlicerHoveredUnselectedItemWithData     = 32
xlSlicerHoveredUnselectedItemWithNoData   = 34
xlSlicerSelectedItemWithData              = 30
xlSlicerSelectedItemWithNoData            = 31
xlSlicerUnselectedItemWithData            = 28
xlSlicerUnselectedItemWithNoData          = 29
xlSubtotalColumn1                         = 13
xlSubtotalColumn2                         = 14
xlSubtotalColumn3                         = 15
xlSubtotalRow1                            = 16
xlSubtotalRow2                            = 17
xlSubtotalRow3                            = 18
xlTimelinePeriodLabels1                   = 38
xlTimelinePeriodLabels2                   = 39
xlTimelineSelectedTimeBlock               = 40
xlTimelineSelectedTimeBlockSpace          = 42
xlTimelineSelectionLabel                  = 36
xlTimelineTimeLevel                       = 37
xlTimelineUnselectedTimeBlock             = 41
xlTotalRow                                = 2
xlWholeTable                              = 0
xlDelimited                               = 1
xlFixedWidth                              = 2
xlTextQualifierDoubleQuote                = 1
xlTextQualifierNone                       = -4142
xlTextQualifierSingleQuote                = 2
xlTextVisualLTR                           = 1
xlTextVisualRTL                           = 2
xlThemeColorAccent1                       = 5
xlThemeColorAccent2                       = 6
xlThemeColorAccent3                       = 7
xlThemeColorAccent4                       = 8
xlThemeColorAccent5                       = 9
xlThemeColorAccent6                       = 10
xlThemeColorDark1                         = 1
xlThemeColorDark2                         = 3
xlThemeColorFollowedHyperlink             = 12
xlThemeColorHyperlink                     = 11
xlThemeColorLight1                        = 2
xlThemeColorLight2                        = 4
xlThemeFontMajor                          = 1
xlThemeFontMinor                          = 2
xlThemeFontNone                           = 0
xlThreadModeAutomatic                     = 0
xlThreadModeManual                        = 1
xlTickLabelOrientationAutomatic           = -4105
xlTickLabelOrientationDownward            = -4170
xlTickLabelOrientationHorizontal          = -4128
xlTickLabelOrientationUpward              = -4171
xlTickLabelOrientationVertical            = -4166
xlTickLabelPositionHigh                   = -4127
xlTickLabelPositionLow                    = -4134
xlTickLabelPositionNextToAxis             = 4
xlTickLabelPositionNone                   = -4142
xlTickMarkCross                           = 4
xlTickMarkInside                          = 2
xlTickMarkNone                            = -4142
xlTickMarkOutside                         = 3
xlLast7Days                               = 2
xlLastMonth                               = 5
xlLastWeek                                = 4
xlNextMonth                               = 8
xlNextWeek                                = 7
xlThisMonth                               = 9
xlThisWeek                                = 3
xlToday                                   = 0
xlTomorrow                                = 6
xlYesterday                               = 1
xlDays                                    = 0
xlMonths                                  = 1
xlYears                                   = 2
xlTimelineLevelDays                       = 3
xlTimelineLevelMonths                     = 2
xlTimelineLevelQuarters                   = 1
xlTimelineLevelYears                      = 0
xlNoButtonChanges                         = 1
xlNoChanges                               = 4
xlNoDockingChanges                        = 3
xlNoShapeChanges                          = 2
xlToolbarProtectionNone                   = -4143
xlTop10Bottom                             = 0
xlTop10Top                                = 1
xlTotalsCalculationAverage                = 2
xlTotalsCalculationCount                  = 3
xlTotalsCalculationCountNums              = 4
xlTotalsCalculationCustom                 = 9
xlTotalsCalculationMax                    = 6
xlTotalsCalculationMin                    = 5
xlTotalsCalculationNone                   = 0
xlTotalsCalculationStdDev                 = 7
xlTotalsCalculationSum                    = 1
xlTotalsCalculationVar                    = 8
xlExponential                             = 5
xlLinear                                  = -4132
xlLogarithmic                             = -4133
xlMovingAvg                               = 6
xlPolynomial                              = 3
xlPower                                   = 4
xlUnderlineStyleDouble                    = -4119
xlUnderlineStyleDoubleAccounting          = 5
xlUnderlineStyleNone                      = -4142
xlUnderlineStyleSingle                    = 2
xlUnderlineStyleSingleAccounting          = 4
xlUpdateLinksAlways                       = 3
xlUpdateLinksNever                        = 2
xlUpdateLinksUserSetting                  = 1
xlVAlignBottom                            = -4107
xlVAlignCenter                            = -4108
xlVAlignDistributed                       = -4117
xlVAlignJustify                           = -4130
xlVAlignTop                               = -4160
xlValueAscending                          = 1
xlValueDescending                         = 2
xlValueNone                               = 0
xlWBATChart                               = -4109
xlWBATExcel4IntlMacroSheet                = 4
xlWBATExcel4MacroSheet                    = 3
xlWBATWorksheet                           = -4167
xlWebFormattingAll                        = 1
xlWebFormattingNone                       = 3
xlWebFormattingRTF                        = 2
xlAllTables                               = 2
xlEntirePage                              = 1
xlSpecifiedTables                         = 3
xlMaximized                               = -4137
xlNormal                                  = -4143
xlMinimized                               = -4140
xlChartAsWindow                           = 5
xlChartInPlace                            = 4
xlClipboard                               = 3
xlInfo                                    = -4129
xlWorkbook                                = 1
xlNormalView                              = 1
xlPageBreakPreview                        = 2
xlPageLayoutView                          = 3
xlCommand                                 = 2
xlFunction                                = 1
xlNotXLM                                  = 3
xlXmlExportSuccess                        = 0
xlXmlExportValidationFailed               = 1
xlXmlImportElementsTruncated              = 1
xlXmlImportSuccess                        = 0
xlXmlImportValidationFailed               = 2
xlXmlLoadImportToList                     = 2
xlXmlLoadMapXml                           = 3
xlXmlLoadOpenXml                          = 1
xlXmlLoadPromptUser                       = 0
xlGuess                                   = 0
xlNo                                      = 2
xlYes                                     = 1
# endregion




"""
import vengeance as ven
import vengeance as vgc
"""
from .conditional import loads_excel_module
from .conditional import read_config_file

read_config_file()

from .version import *
from .util    import *
from .classes import *

if loads_excel_module:
    from .excel_com import *

del loads_excel_module
del read_config_file






